from __future__ import annotations

import ast
import os
from collections.abc import Iterable
from configparser import ConfigParser
from dataclasses import asdict, dataclass, field, fields
from importlib.metadata import Distribution
from pathlib import Path
from typing import TYPE_CHECKING, Any, no_type_check

if TYPE_CHECKING:
    from importlib.metadata import _SimplePath


@dataclass
class Setup:
    """
    Abstraction of a Python project setup file.
    """

    name: str | None = None
    version: str | None = None
    install_requires: list[str] = field(default_factory=list)
    extras_require: dict[str, list[str]] = field(default_factory=dict)
    python_requires: str | None = None
    summary: str | None = None

    def update(self, other: Setup) -> None:
        for f in fields(self):
            other_field = getattr(other, f.name)
            if other_field:
                setattr(self, f.name, other_field)

    def as_dict(self) -> dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_directory(cls, dir: Path) -> Setup:
        return _SetupReader.read_from_directory(dir)

    def as_dist(self) -> Distribution:
        return SetupDistribution(self)


class _SetupReader:
    """
    Class that reads a setup.py file without executing it.
    """

    @classmethod
    def read_from_directory(cls, directory: Path) -> Setup:
        result = Setup()

        for filename, file_reader in [
            ("pyproject.toml", cls.read_pyproject_toml),
            ("setup.cfg", cls.read_setup_cfg),
            ("setup.py", cls.read_setup_py),
        ]:
            filepath = directory / filename
            if not filepath.exists():
                continue

            new_result = file_reader(filepath)
            result.update(new_result)

        return result

    @staticmethod
    def read_pyproject_toml(file: Path) -> Setup:
        from pdm import termui
        from pdm.exceptions import ProjectError
        from pdm.formats import MetaConvertError
        from pdm.project.project_file import PyProject

        try:
            metadata = PyProject(file, ui=termui.UI()).metadata
        except ProjectError:
            return Setup()
        except MetaConvertError as e:
            termui.logger.warning("Error parsing pyproject.toml, metadata may be incomplete. %s", e)
            metadata = e.data
        return Setup(
            name=metadata.get("name"),
            summary=metadata.get("description"),
            version=metadata.get("version"),
            install_requires=metadata.get("dependencies", []),
            extras_require=metadata.get("optional-dependencies", {}),
            python_requires=metadata.get("requires-python"),
        )

    @no_type_check
    @classmethod
    def read_setup_py(cls, file: Path) -> Setup:
        with file.open(encoding="utf-8") as f:
            content = f.read()

        body = ast.parse(content).body

        setup_call, body = cls._find_setup_call(body)
        if not setup_call:
            return Setup()

        return Setup(
            name=cls._find_single_string(setup_call, body, "name"),
            version=cls._find_single_string(setup_call, body, "version") or "0.0.0",
            install_requires=cls._find_install_requires(setup_call, body),
            extras_require=cls._find_extras_require(setup_call, body),
            python_requires=cls._find_single_string(setup_call, body, "python_requires"),
        )

    @staticmethod
    def read_setup_cfg(file: Path) -> Setup:
        parser = ConfigParser()

        parser.read(str(file))

        name = None
        version = "0.0.0"
        if parser.has_option("metadata", "name"):
            name = parser.get("metadata", "name")

        if parser.has_option("metadata", "version"):
            meta_version = parser.get("metadata", "version")
            if not meta_version.startswith("attr:"):
                version = meta_version

        install_requires = []
        extras_require: dict[str, list[str]] = {}
        python_requires = None
        if parser.has_section("options"):
            if parser.has_option("options", "install_requires"):
                for dep in parser.get("options", "install_requires").split("\n"):
                    dep = dep.strip()
                    if not dep:
                        continue

                    install_requires.append(dep)

            if parser.has_option("options", "python_requires"):
                python_requires = parser.get("options", "python_requires")

        if parser.has_section("options.extras_require"):
            for group in parser.options("options.extras_require"):
                extras_require[group] = []
                deps = parser.get("options.extras_require", group)
                for dep in deps.split("\n"):
                    dep = dep.strip()
                    if not dep:
                        continue

                    extras_require[group].append(dep)

        return Setup(
            name=name,
            version=version,
            install_requires=install_requires,
            extras_require=extras_require,
            python_requires=python_requires,
        )

    @classmethod
    def _find_setup_call(cls, elements: list[Any]) -> tuple[ast.Call | None, list[Any | None]]:
        funcdefs = []
        for i, element in enumerate(elements):
            if isinstance(element, ast.If) and i == len(elements) - 1:
                # Checking if the last element is an if statement
                # and if it is 'if __name__ == "__main__"' which
                # could contain the call to setup()
                test = element.test
                if not isinstance(test, ast.Compare):
                    continue

                left = test.left
                if not isinstance(left, ast.Name):
                    continue

                if left.id != "__name__":
                    continue

                setup_call, body = cls._find_sub_setup_call([element])
                if not setup_call:
                    continue

                return setup_call, body + elements
            if not isinstance(element, ast.Expr):
                if isinstance(element, ast.FunctionDef):
                    funcdefs.append(element)

                continue

            value = element.value
            if not isinstance(value, ast.Call):
                continue

            func = value.func
            if not (isinstance(func, ast.Name) and func.id == "setup") and not (
                isinstance(func, ast.Attribute)
                and isinstance(func.value, ast.Name)
                and func.value.id == "setuptools"
                and func.attr == "setup"
            ):
                continue

            return value, elements

        # Nothing, we inspect the function definitions
        return cls._find_sub_setup_call(funcdefs)

    @no_type_check
    @classmethod
    def _find_sub_setup_call(cls, elements: list[Any]) -> tuple[ast.Call | None, list[Any | None]]:
        for element in elements:
            if not isinstance(element, (ast.FunctionDef, ast.If)):
                continue

            setup_call = cls._find_setup_call(element.body)
            if setup_call != (None, None):
                setup_call, body = setup_call

                body = elements + body

                return setup_call, body

        return None, None

    @no_type_check
    @classmethod
    def _find_install_requires(cls, call: ast.Call, body: Iterable[Any]) -> list[str]:
        install_requires: list[str] = []
        value = cls._find_in_call(call, "install_requires")
        if value is None:
            # Trying to find in kwargs
            kwargs = cls._find_call_kwargs(call)

            if kwargs is None or not isinstance(kwargs, ast.Name):
                return install_requires

            variable = cls._find_variable_in_body(body, kwargs.id)
            if not isinstance(variable, (ast.Dict, ast.Call)):
                return install_requires

            if isinstance(variable, ast.Call):
                if not isinstance(variable.func, ast.Name):
                    return install_requires

                if variable.func.id != "dict":
                    return install_requires

                value = cls._find_in_call(variable, "install_requires")
            else:
                value = cls._find_in_dict(variable, "install_requires")

        if value is None:
            return install_requires

        if isinstance(value, ast.List):
            install_requires.extend(
                [el.value for el in value.elts if isinstance(el, ast.Constant) and isinstance(el.value, str)]
            )
        elif isinstance(value, ast.Name):
            variable = cls._find_variable_in_body(body, value.id)

            if variable is not None and isinstance(variable, ast.List):
                install_requires.extend(
                    [el.value for el in variable.elts if isinstance(el, ast.Constant) and isinstance(el.value, str)]
                )

        return install_requires

    @no_type_check
    @classmethod
    def _find_extras_require(cls, call: ast.Call, body: Iterable[Any]) -> dict[str, list[str]]:
        extras_require: dict[str, list[str]] = {}
        value = cls._find_in_call(call, "extras_require")
        if value is None:
            # Trying to find in kwargs
            kwargs = cls._find_call_kwargs(call)

            if kwargs is None or not isinstance(kwargs, ast.Name):
                return extras_require

            variable = cls._find_variable_in_body(body, kwargs.id)
            if not isinstance(variable, (ast.Dict, ast.Call)):
                return extras_require

            if isinstance(variable, ast.Call):
                if not isinstance(variable.func, ast.Name):
                    return extras_require

                if variable.func.id != "dict":
                    return extras_require

                value = cls._find_in_call(variable, "extras_require")
            else:
                value = cls._find_in_dict(variable, "extras_require")

        if value is None:
            return extras_require

        if isinstance(value, ast.Dict):
            for key, val in zip(value.keys, value.values, strict=True):
                if isinstance(val, ast.Name):
                    val = cls._find_variable_in_body(body, val.id)

                if isinstance(val, ast.List):
                    extras_require[key.value] = [
                        e.value for e in val.elts if isinstance(e, ast.Constant) and isinstance(e.value, str)
                    ]
        elif isinstance(value, ast.Name):
            variable = cls._find_variable_in_body(body, value.id)

            if variable is None or not isinstance(variable, ast.Dict):
                return extras_require

            for key, val in zip(variable.keys, variable.values, strict=True):
                if isinstance(val, ast.Name):
                    val = cls._find_variable_in_body(body, val.id)

                if isinstance(val, ast.List):
                    extras_require[key.value] = [
                        e.value for e in val.elts if isinstance(e, ast.Constant) and isinstance(e.value, str)
                    ]

        return extras_require

    @classmethod
    def _find_single_string(cls, call: ast.Call, body: list[Any], name: str) -> str | None:
        value = cls._find_in_call(call, name)
        if value is None:
            # Trying to find in kwargs
            kwargs = cls._find_call_kwargs(call)

            if kwargs is None or not isinstance(kwargs, ast.Name):
                return None

            variable = cls._find_variable_in_body(body, kwargs.id)
            if not isinstance(variable, (ast.Dict, ast.Call)):
                return None

            if isinstance(variable, ast.Call):
                if not isinstance(variable.func, ast.Name):
                    return None

                if variable.func.id != "dict":
                    return None

                value = cls._find_in_call(variable, name)
            else:
                value = cls._find_in_dict(variable, name)

        if value is None:
            return None

        if isinstance(value, ast.Constant) and isinstance(value.value, str):
            return value.value
        elif isinstance(value, ast.Name):
            variable = cls._find_variable_in_body(body, value.id)

            if variable is not None and isinstance(variable, ast.Constant) and isinstance(variable.value, str):
                return variable.value

        return None

    @staticmethod
    def _find_in_call(call: ast.Call, name: str) -> Any | None:
        for keyword in call.keywords:
            if keyword.arg == name:
                return keyword.value
        return None

    @staticmethod
    def _find_call_kwargs(call: ast.Call) -> Any | None:
        kwargs = None
        for keyword in call.keywords:
            if keyword.arg is None:
                kwargs = keyword.value

        return kwargs

    @staticmethod
    def _find_variable_in_body(body: Iterable[Any], name: str) -> Any | None:
        for elem in body:
            if not isinstance(elem, ast.Assign):
                continue

            for target in elem.targets:
                if not isinstance(target, ast.Name):
                    continue

                if target.id == name:
                    return elem.value
        return None

    @staticmethod
    def _find_in_dict(dict_: ast.Dict, name: str) -> Any | None:
        for key, val in zip(dict_.keys, dict_.values, strict=True):
            if isinstance(key, ast.Constant) and key.value == name:
                return val
        return None


class SetupDistribution(Distribution):
    def __init__(self, data: Setup) -> None:
        self._data = data

    def read_text(self, filename: str) -> str | None:
        return None

    def locate_file(self, path: str | os.PathLike[str]) -> _SimplePath:
        return Path()

    @property
    def metadata(self) -> dict[str, Any]:  # type: ignore[override]
        return {
            k: v
            for k, v in {
                "Name": self._data.name,
                "Version": self._data.version,
                "Summary": self._data.summary,
                "Requires-Python": self._data.python_requires,
            }.items()
            if v is not None
        }

    @property
    def requires(self) -> list[str] | None:
        from pdm.models.markers import get_marker
        from pdm.models.requirements import parse_requirement

        result = self._data.install_requires[:]
        for extra, reqs in self._data.extras_require.items():
            extra_marker = f"extra == '{extra}'"
            for req in reqs:
                parsed = parse_requirement(req)
                old_marker = str(parsed.marker) if parsed.marker else None
                if old_marker:
                    if " or " in old_marker:
                        new_marker = f"({old_marker}) and {extra_marker}"
                    else:
                        new_marker = f"{old_marker} and {extra_marker}"
                else:
                    new_marker = extra_marker
                parsed.marker = get_marker(new_marker)
                result.append(parsed.as_line())
        return result



from pathlib import Path
from unittest.mock import ANY

import pytest
from unearth import Link

from pdm.cli import actions
from pdm.exceptions import PdmUsageError
from pdm.models.requirements import parse_requirement
from pdm.models.specifiers import PySpecSet
from pdm.project.lockfile import FLAG_CROSS_PLATFORM
from pdm.project.lockfile.base import Compatibility
from pdm.signals import pre_lock
from pdm.utils import parse_version
from tests import FIXTURES


def test_lock_command(project, pdm, mocker):
    m = mocker.patch.object(actions, "do_lock")
    pdm(["lock"], obj=project)
    m.assert_called_with(
        project,
        refresh=False,
        groups=["default"],
        hooks=ANY,
        strategy_change=None,
        strategy="all",
        append=False,
        env_spec=None,
    )


class FakeResolver:
    def __init__(self, *args, **kwargs):
        pass

    def resolve(self):
        return [], set()


def make_workspace_member(project, core):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    return core.create_project(member_path, global_config=project.global_config.config_file.as_posix())


def make_workspace_members(project, core, members):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    for name, dependencies in members.items():
        member_path = project.root / "packages" / name
        member_path.mkdir(parents=True)
        dependency_lines = ", ".join(f'"{dependency}"' for dependency in dependencies)
        member_path.joinpath("pyproject.toml").write_text(
            f'[project]\nname = "{name}"\nversion = "0.1.0"\ndependencies = [{dependency_lines}]\n',
            encoding="utf-8",
        )
    return core.create_project(project.root, global_config=project.global_config.config_file.as_posix())


def capture_pre_lock_requirements():
    captured = {}

    def capture_requirements(sender, requirements, **kwargs):
        captured["requirements"] = requirements

    pre_lock.connect(capture_requirements, weak=False)
    return captured, capture_requirements


def test_do_lock_adds_workspace_members_to_explicit_member_requirements(project, core, mocker):
    member_project = make_workspace_member(project, core)
    mocker.patch.object(member_project, "get_resolver", return_value=FakeResolver)
    captured, receiver = capture_pre_lock_requirements()

    try:
        actions.do_lock(
            member_project,
            requirements=[parse_requirement("requests")],
            groups=["default"],
            dry_run=True,
        )
    finally:
        pre_lock.disconnect(receiver)

    requirements = captured["requirements"]
    assert [req.identify() for req in requirements] == ["requests", "foo"]
    assert requirements[-1].editable
    assert requirements[-1].str_path == "./packages/foo"
    assert member_project.lockfile._path == project.root / "pdm.lock"


def test_do_lock_adds_workspace_members_to_member_resolved_requirements(project, core, mocker):
    member_project = make_workspace_member(project, core)
    mocker.patch.object(member_project, "get_resolver", return_value=FakeResolver)
    captured, receiver = capture_pre_lock_requirements()

    try:
        actions.do_lock(member_project, groups=["default"], dry_run=True)
    finally:
        pre_lock.disconnect(receiver)

    requirements = captured["requirements"]
    assert [req.identify() for req in requirements] == ["requests", "foo"]
    assert requirements[-1].editable
    assert requirements[-1].str_path == "./packages/foo"
    assert member_project.lockfile._path == project.root / "pdm.lock"


def test_lock_workspace_members_depend_on_each_other(project, core, repository):
    workspace_project = make_workspace_members(project, core, {"foo": ["bar"], "bar": []})

    actions.do_lock(workspace_project)

    locked_candidates = workspace_project.get_locked_repository().candidates
    assert locked_candidates["foo"].req.editable
    assert locked_candidates["bar"].req.editable
    assert locked_candidates["foo"].req.str_path == "./packages/foo"
    assert locked_candidates["bar"].req.str_path == "./packages/bar"


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_hash_tracks_workspace_member_pyproject(project, core, lock_format):
    project.project_config["lock.format"] = lock_format
    workspace_project = make_workspace_members(project, core, {"foo": []})

    actions.do_lock(workspace_project)

    assert workspace_project.is_lockfile_hash_match()

    member_pyproject = workspace_project.root / "packages" / "foo" / "pyproject.toml"
    member_pyproject.write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    fresh_project = core.create_project(
        workspace_project.root, global_config=workspace_project.global_config.config_file.as_posix()
    )
    fresh_member_project = core.create_project(
        member_pyproject.parent, global_config=workspace_project.global_config.config_file.as_posix()
    )

    assert not fresh_project.is_lockfile_hash_match()
    assert not fresh_member_project.is_lockfile_hash_match()

    actions.do_lock(fresh_project)

    assert fresh_project.is_lockfile_hash_match()


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("lock_format", ["pdm", "pylock"])
def test_lock_dependencies(project, lock_format):
    project.add_dependencies(["requests"])
    project.project_config["lock.format"] = lock_format
    actions.do_lock(project)
    assert project.lockfile.exists()
    assert project.lockfile._path.name == "pdm.lock" if lock_format == "pdm" else "pylock.toml"
    locked = project.get_locked_repository().candidates
    for package in ("requests", "idna", "chardet", "certifi"):
        assert package in locked


@pytest.mark.parametrize("args", [("-S", "static_urls"), ("--static-urls",)])
def test_lock_refresh(pdm, project, repository, args, core, mocker):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert not package.get("files")
    project.add_dependencies(["requests>=2.0"])
    url_hashes = {
        "http://example.com/requests-2.19.1-py3-none-any.whl": "sha256:abcdef123456",
        "http://example2.com/requests-2.19.1-py3-none-AMD64.whl": "sha256:abcdef123456",
        "http://example1.com/requests-2.19.1-py3-none-any.whl": "sha256:abcdef123456",
    }
    mocker.patch.object(
        core.repository_class,
        "get_hashes",
        side_effect=(
            lambda c: (
                [{"url": url, "file": Link(url).filename, "hash": hash} for url, hash in url_hashes.items()]
                if c.identify() == "requests"
                else []
            )
        ),
    )
    assert not project.is_lockfile_hash_match()
    result = pdm(["lock", "--refresh", "-v"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert package["files"] == [
        {"file": "requests-2.19.1-py3-none-AMD64.whl", "hash": "sha256:abcdef123456"},
        {"file": "requests-2.19.1-py3-none-any.whl", "hash": "sha256:abcdef123456"},
    ]
    assert project.is_lockfile_hash_match()
    result = pdm(["lock", "--refresh", *args, "-v"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["package"] if p["name"] == "requests")
    assert package["files"] == [{"url": url, "hash": hash} for url, hash in sorted(url_hashes.items())]


def test_lock_refresh_keep_consistent(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()
    previous = project.lockfile._path.read_text()
    result = pdm(["lock", "--refresh"], obj=project)
    assert result.exit_code == 0
    assert project.lockfile._path.read_text() == previous


def test_pylock_refresh_preserves_dependency_group_markers(pdm, project, repository):
    project.project_config["lock.format"] = "pylock"
    project.add_dependencies(["requests"], to_group="dev", dev=True)
    result = pdm(["lock", "-G", "dev"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["packages"] if p["name"] == "requests")
    assert package["marker"] == '"dev" in dependency_groups'

    result = pdm(["lock", "--refresh"], obj=project)
    assert result.exit_code == 0
    package = next(p for p in project.lockfile["packages"] if p["name"] == "requests")
    assert package["marker"] == '"dev" in dependency_groups'


def test_lock_check_no_change_success(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()

    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 0


def test_lock_check_change_fails(pdm, project, repository):
    project.add_dependencies(["requests"])
    result = pdm(["lock"], obj=project)
    assert result.exit_code == 0
    assert project.is_lockfile_hash_match()

    project.add_dependencies(["pyyaml"])
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == 1


@pytest.mark.usefixtures("repository")
def test_innovations_with_specified_lockfile(pdm, project, working_set):
    project.add_dependencies(["requests"])
    lockfile = str(project.root / "mylock.lock")
    pdm(["lock", "--lockfile", lockfile], strict=True, obj=project)
    assert project.lockfile._path == project.root / "mylock.lock"
    assert project.is_lockfile_hash_match()
    locked = project.get_locked_repository().candidates
    assert "requests" in locked
    pdm(["sync", "--lockfile", lockfile], strict=True, obj=project)
    assert "requests" in working_set


@pytest.mark.usefixtures("repository", "vcs")
def test_skip_editable_dependencies_in_metadata(project, capsys):
    project.pyproject.metadata["dependencies"] = [
        "-e git+https://github.com/test-root/demo.git@1234567890abcdef#egg=demo"
    ]
    actions.do_lock(project)
    _, err = capsys.readouterr()
    assert "WARNING: Skipping editable dependency" in err
    assert not project.get_locked_repository().candidates


@pytest.mark.usefixtures("repository")
def test_lock_selected_groups(project, pdm):
    project.add_dependencies(["requests"], to_group="http")
    project.add_dependencies(["pytz"])
    pdm(["lock", "-G", "http", "--no-default"], obj=project, strict=True)
    assert project.lockfile.groups == ["http"]
    assert "requests" in project.get_locked_repository().candidates
    assert "pytz" not in project.get_locked_repository().candidates


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("to_dev", [True, False])
def test_lock_self_referencing_dev_groups(project, pdm, to_dev):
    name = project.name
    project.add_dependencies(["requests"], to_group="http", dev=to_dev)
    project.add_dependencies(
        {"pytz": parse_requirement("pytz"), f"{name}[http]": parse_requirement(f"{name}[http]")},
        to_group="dev",
        dev=True,
    )
    pdm(["lock", "-G", "dev"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "dev", "http"]
    packages = project.lockfile["package"]
    pytz = next(p for p in packages if p["name"] == "pytz")
    assert pytz["groups"] == ["dev"]
    requests = next(p for p in packages if p["name"] == "requests")
    assert requests["groups"] == ["dev", "http"]
    idna = next(p for p in packages if p["name"] == "idna")
    assert idna["groups"] == ["dev", "http"]


@pytest.mark.usefixtures("repository")
def test_lock_self_referencing_optional_groups(project, pdm):
    name = project.name
    project.add_dependencies(["requests"], to_group="http")
    project.add_dependencies(
        {"pytz": parse_requirement("pytz"), f"{name}[http]": parse_requirement(f"{name}[http]")},
        to_group="all",
    )
    pdm(["lock", "-G", "all"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "all", "http"]
    packages = project.lockfile["package"]
    pytz = next(p for p in packages if p["name"] == "pytz")
    assert pytz["groups"] == ["all"]
    requests = next(p for p in packages if p["name"] == "requests")
    assert requests["groups"] == ["all", "http"]
    idna = next(p for p in packages if p["name"] == "idna")
    assert idna["groups"] == ["all", "http"]


@pytest.mark.usefixtures("repository")
def test_lock_include_groups_not_allowed(project, pdm):
    project.pyproject.metadata["optional-dependencies"] = {"http": ["requests"]}
    project.pyproject.dependency_groups.update({"dev": ["pytest", {"include-group": "http"}]})
    project.pyproject.write()
    result = pdm(["lock", "-G", "all"], obj=project)
    assert result.exit_code != 0
    assert "Missing group 'http' in `include-group`" in result.stderr


@pytest.mark.usefixtures("repository")
def test_lock_optional_referencing_dev_group_not_allowed(project, pdm):
    name = project.name
    project.pyproject.metadata["optional-dependencies"] = {"http": ["requests", f"{name}[dev]"]}
    project.pyproject.dependency_groups.update({"dev": ["pytest"]})
    project.pyproject.write()
    result = pdm(["lock", "-G", "http"], obj=project)
    assert result.exit_code != 0
    assert "Optional dependency group 'http' cannot include non-existing extras" in result.stderr


@pytest.mark.usefixtures("local_finder")
def test_lock_multiple_platform_wheels(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.7")
    project.add_dependencies(["pdm-hello"])
    pdm(["lock"], obj=project, strict=True)
    package = next(p for p in project.lockfile["package"] if p["name"] == "pdm-hello")
    file_hashes = package["files"]
    assert len(file_hashes) == 2


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize("platform", ["linux", "macos", "windows"])
def test_lock_specific_platform_wheels(project, pdm, platform):
    project.environment.python_requires = PySpecSet(">=3.7")
    project.add_dependencies(["pdm-hello"])
    pdm(["lock", "--platform", platform], obj=project, strict=True)
    assert FLAG_CROSS_PLATFORM not in project.lockfile.strategy
    package = next(p for p in project.lockfile["package"] if p["name"] == "pdm-hello")
    file_hashes = package["files"]
    wheels_num = 2 if platform == "windows" else 1
    assert len(file_hashes) == wheels_num


def test_parse_lock_strategy_group_options(core):
    core.init_parser()
    parser = core.parser

    ns = parser.parse_args(["lock", "-S", "no_cross_platform"])
    assert ns.strategy_change == ["no_cross_platform"]
    ns = parser.parse_args(["lock", "-S", "no_cross_platform", "--static-urls"])
    assert ns.strategy_change == ["no_cross_platform", "static_urls"]
    ns = parser.parse_args(["lock", "-S", "no_cross_platform,direct_minimal_versions"])
    assert ns.strategy_change == ["no_cross_platform", "direct_minimal_versions"]


def test_apply_lock_strategy_changes(project):
    assert project.lockfile.apply_strategy_change(["no_cross_platform", "static_urls"]) == {
        "inherit_metadata",
        "static_urls",
    }
    assert project.lockfile.apply_strategy_change(["no_static_urls"]) == {"inherit_metadata"}
    assert project.lockfile.apply_strategy_change(["no_inherit_metadata"]) == set()


@pytest.mark.parametrize("strategy", [["abc"], ["no_abc", "static_urls"]])
def test_apply_lock_strategy_changes_invalid(project, strategy):
    with pytest.raises(PdmUsageError):
        project.lockfile.apply_strategy_change(strategy)


def test_lock_direct_minimal_versions(project, repository, pdm):
    project.add_dependencies(["django"])
    repository.add_candidate("pytz", "2019.6")
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    assert project.lockfile.strategy == {"direct_minimal_versions", "inherit_metadata"}
    locked_repository = project.get_locked_repository()
    assert locked_repository.candidates["django"].version == "1.11.8"
    assert locked_repository.candidates["pytz"].version == "2019.6"


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize("args", [(), ("-S", "direct_minimal_versions")])
def test_lock_direct_minimal_versions_real(project, pdm, args):
    project.add_dependencies(["zipp"])
    pdm(["lock", *args], obj=project, strict=True)
    locked_candidate = project.get_locked_repository().candidates["zipp"]
    if args:
        assert locked_candidate.version == "3.6.0"
    else:
        assert locked_candidate.version == "3.7.0"


@pytest.mark.parametrize(
    "lock_version,expected",
    [
        ("4.1.0", Compatibility.BACKWARD),
        ("4.1.1", Compatibility.SAME),
        ("4.1.2", Compatibility.FORWARD),
        ("4.2", Compatibility.NONE),
        ("3.0", Compatibility.NONE),
        ("4.0.1", Compatibility.BACKWARD),
    ],
)
def test_lockfile_compatibility(project, monkeypatch, lock_version, expected, pdm):
    pdm(["lock"], obj=project, strict=True)
    monkeypatch.setattr("pdm.project.lockfile.PDMLock.spec_version", parse_version("4.1.1"))
    project.lockfile._data["metadata"]["lock_version"] = lock_version
    assert project.lockfile.compatibility() == expected
    result = pdm(["lock", "--check"], obj=project)
    assert result.exit_code == (1 if expected == Compatibility.NONE else 0)


def test_lock_default_inherit_metadata(project, pdm, mocker, working_set):
    project.add_dependencies(["requests"])
    pdm(["lock"], obj=project, strict=True)
    assert "inherit_metadata" in project.lockfile.strategy
    packages = project.lockfile["package"]
    assert all(package["groups"] == ["default"] for package in packages)

    resolver = mocker.patch.object(project, "get_resolver")
    pdm(["sync"], obj=project, strict=True)
    resolver.assert_not_called()
    for key in ("requests", "idna", "chardet", "urllib3"):
        assert key in working_set


def test_lock_inherit_metadata_strategy(project, pdm, mocker, working_set):
    project.add_dependencies(["requests"])
    pdm(["lock", "-S", "inherit_metadata"], obj=project, strict=True)
    assert "inherit_metadata" in project.lockfile.strategy
    packages = project.lockfile["package"]
    assert all(package["groups"] == ["default"] for package in packages)

    resolver = mocker.patch.object(project, "get_resolver")
    pdm(["sync"], obj=project, strict=True)
    resolver.assert_not_called()
    for key in ("requests", "idna", "chardet", "urllib3"):
        assert key in working_set


@pytest.mark.parametrize("source", ["cli", "pyproject", "config"])
def test_lock_exclude_newer(project, pdm, source):
    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.add_dependencies(["zipp"])
    pdm(["lock"], obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.7.0"

    cmd = ["lock", "--exclude-newer", "2024-01-01"]
    if source == "pyproject":
        project.pyproject.settings["resolution"] = {"exclude-newer": "2024-01-01"}
        cmd = ["lock"]
    elif source == "config":
        project.project_config["strategy.exclude-newer"] = "2024-01-01"
        cmd = ["lock"]

    pdm(cmd, obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.6.0"


@pytest.mark.parametrize("source", ["cli", "pyproject", "config"])
def test_lock_exclude_newer_accepts_relative_duration(project, pdm, monkeypatch, source):
    from datetime import datetime, timezone

    import pdm.utils as pdm_utils

    frozen_now = datetime(2024, 2, 9, 12, 0, tzinfo=timezone.utc)

    class FrozenDateTime(datetime):
        @classmethod
        def now(cls, tz=None):
            assert tz is timezone.utc
            return frozen_now

    monkeypatch.setattr(pdm_utils, "datetime", FrozenDateTime)

    cmd = ["lock", "--exclude-newer", "3w"]
    if source == "pyproject":
        project.pyproject.settings["resolution"] = {"exclude-newer": "3w"}
        cmd = ["lock"]
    elif source == "config":
        project.project_config["strategy.exclude-newer"] = "3w"
        cmd = ["lock"]

    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.add_dependencies(["zipp"])
    pdm(cmd, strict=True, obj=project, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.6.0"


def test_pyproject_exclude_newer_overrides_config(project, pdm):
    project.pyproject.metadata["requires-python"] = ">=3.9"
    project.project_config["pypi.url"] = "https://my.pypi.org/json"
    project.project_config["strategy.exclude-newer"] = "2024-01-01"
    project.pyproject.settings["resolution"] = {"exclude-newer": "2025-01-01"}
    project.add_dependencies(["zipp"])

    pdm(["lock"], obj=project, strict=True, cleanup=False)
    assert project.get_locked_repository().candidates["zipp"].version == "3.7.0"


exclusion_cases = [
    pytest.param(("-G", ":all", "--without", "tz,ssl"), id="-G :all --without tz,ssl"),
    pytest.param(("-G", ":all", "--without", "tz", "--without", "ssl"), id="-G :all --without tz --without ssl"),
    pytest.param(("--with", ":all", "--without", "tz,ssl"), id="--with all --without tz,ssl"),
    pytest.param(("--with", ":all", "--without", "tz", "--without", "ssl"), id="--with all --without tz --without ssl"),
    pytest.param(("--without", "tz", "--without", "ssl"), id="--without tz --without ssl"),
    pytest.param(("--without", "tz,ssl"), id="--without tz,ssl"),
]


@pytest.mark.parametrize("args", exclusion_cases)
@pytest.mark.usefixtures("repository")
def test_lock_all_with_excluded_groups(project, pdm, args):
    project.add_dependencies(["urllib3"], "url")
    project.add_dependencies(["pytz"], "tz", True)
    project.add_dependencies(["pyopenssl"], "ssl")
    pdm(["lock", *args], obj=project, strict=True)
    assert "urllib3" in project.get_locked_repository().candidates
    assert "pytz" not in project.get_locked_repository().candidates
    assert "pyopenssl" not in project.get_locked_repository().candidates


@pytest.mark.parametrize(
    "args",
    [
        ("--append",),
        ("--python", "<3.6"),
        ("-S", "cross_platform", "--append", "--python", "3.10"),
        ("--platform", "linux", "--refresh"),
    ],
)
def test_forbidden_lock_target_options(project, pdm, args):
    result = pdm(["lock", *args], obj=project)
    assert result.exit_code != 0
    assert "PdmUsageError" in result.stderr


@pytest.mark.parametrize("nested", [False, True])
def test_lock_for_multiple_targets(project, pdm, repository, nested):
    deps = [
        'django<2; sys_platform == "win32"',
        'django>=2; sys_platform != "win32"',
    ]
    if nested:
        repository.add_candidate("foo", "0.1.0")
        repository.add_dependencies("foo", "0.1.0", deps)
        project.add_dependencies(["foo"])
    else:
        project.add_dependencies(deps)

    pdm(["lock", "--platform", "windows"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(candidates["django"]) == 1
    assert candidates["django"][0].version == "1.11.8"
    assert len(locked.targets) == 1
    pytz = candidates["pytz"][0]
    assert str(pytz.req.marker) == 'sys_platform == "win32"'

    result = pdm(["lock", "--platform", "windows", "--append"], obj=project, strict=True)
    assert "already exists, skip locking." in result.stdout

    pdm(["lock", "--platform", "linux", "--append"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(locked.targets) == 2
    assert sorted(c.version for c in candidates["django"]) == ["1.11.8", "2.2.9"]
    pytz = candidates["pytz"][0]
    assert not pytz.req.marker or pytz.req.marker.is_any()

    # not append but overwrite
    pdm(["lock", "--platform", "windows"], obj=project, strict=True)
    locked = project.get_locked_repository()
    candidates = locked.all_candidates
    assert len(candidates["django"]) == 1
    assert candidates["django"][0].version == "1.11.8"
    assert len(locked.targets) == 1
    pytz = candidates["pytz"][0]
    assert str(pytz.req.marker) == 'sys_platform == "win32"'


CONSTRAINT_FILE = str(FIXTURES / "constraints.txt")


@pytest.mark.usefixtures("repository")
@pytest.mark.parametrize("constraint", [CONSTRAINT_FILE, Path(CONSTRAINT_FILE).as_uri()])
def test_lock_with_override_file(project, pdm, constraint):
    project.add_dependencies(["requests"])
    pdm(["lock", "--override", constraint], obj=project, strict=True)
    candidates = project.get_locked_repository().candidates
    assert candidates["requests"].version == "2.20.0b1"
    assert candidates["urllib3"].version == "1.23b0"
    assert "django" not in candidates


def test_pylock_add_remove_strategy(project, pdm):
    project.project_config["lock.format"] = "pylock"
    pdm(["lock"], obj=project, strict=True)
    assert project.lockfile.strategy == {"inherit_metadata", "static_urls"}
    pdm(["lock", "-S", "static_urls"], obj=project, strict=True)
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    assert project.lockfile.strategy == {"inherit_metadata", "static_urls", "direct_minimal_versions"}

    result = pdm(["lock", "-S", "no_static_urls"], obj=project)
    assert result.exit_code != 0
    result = pdm(["lock", "-S", "no_inherit_metadata"], obj=project)
    assert result.exit_code != 0


@pytest.mark.usefixtures("repository")
def test_lock_with_invalid_python_requirement(project, pdm):
    project.add_dependencies(["requests", "python>=3.6"])
    result = pdm(["lock", "-v"], obj=project, strict=True)
    assert "requests" in project.get_locked_repository().candidates
    assert "python" not in project.get_locked_repository().candidates
    assert "The 'python' requirement is not necessary and will be ignored." in result.stderr



from setuptools import setup

if True:
    raise RuntimeError("This mimics the build error on unmatched platform")

setup(
    name="demo",
    version="0.0.1",
    description="test demo",
    py_modules=["demo"],
    python_requires=">=3.3",
)



import argparse
import sys
from textwrap import dedent
from unittest.mock import ANY

import pytest

from pdm.cli.commands.init import Command
from pdm.compat import tomllib
from pdm.models.python import PythonInfo
from pdm.utils import cd

PYTHON_VERSION = f"{sys.version_info[0]}.{sys.version_info[1]}"


@pytest.fixture(autouse=True)
def enable_interactive_mode(mocker):
    from rich import get_console

    console = get_console()
    mocker.patch.object(console, "is_interactive", True)


def test_init_validate_python_requires(project_no_init, pdm):
    result = pdm(["init"], input="\n\n\n\n\n\n\n3.7\n", obj=project_no_init)
    assert result.exit_code != 0
    assert "InvalidSpecifier" in result.stderr


def test_init_command(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    pdm(["init"], input="\ntest-project\n\n\n\n\n\n\n\n", strict=True, obj=project_no_init)
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_uses_existing_pyproject_values_as_defaults(project_no_init, mocker):
    project_no_init.pyproject._path.write_text(
        dedent(
            """
            [project]
            name = "existing-project"
            version = "2.1.0"
            description = "Existing description"
            authors = [{name = "Existing Author", email = "author@example.org"}]
            license = {text = "Apache-2.0"}
            requires-python = ">=3.10"

            [build-system]
            requires = ["hatchling"]
            build-backend = "hatchling.build"

            [tool.pdm]
            distribution = true
            """
        )
    )
    project_no_init.pyproject.reload()
    defaults = {}

    def ask_with_default(question, default, **kwargs):
        defaults[question] = default
        return default

    mocker.patch("pdm.cli.commands.init.termui.ask", side_effect=ask_with_default)
    confirm = mocker.patch(
        "pdm.cli.commands.init.termui.confirm", side_effect=lambda *args, **kwargs: kwargs["default"]
    )
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Git User", "git@example.org"),
    )
    command = Command()
    options = argparse.Namespace(
        name=None,
        project_version=None,
        dist=False,
        backend=None,
        license=None,
    )

    command.get_metadata_from_input(project_no_init, options)

    assert defaults == {
        "Project name": "existing-project",
        "Project version": "2.1.0",
        "Project description": "Existing description",
        "Please select": 3,
        "License(SPDX name)": "Apache-2.0",
        "Author name": "Existing Author",
        "Author email": "author@example.org",
        "Python requires('*' to allow any)": ">=3.10",
    }
    confirm.assert_called_once_with(
        "Do you want to build this project for distribution(such as wheel)?\n"
        "If yes, it will be installed by default when running `pdm install`.",
        default=True,
    )


def test_init_preserves_existing_pyproject_values(project_no_init, pdm, mocker):
    project_no_init.pyproject._path.write_text(
        dedent(
            """
            [project]
            name = "existing-project"
            version = "2.1.0"
            description = "Existing description"
            authors = [{name = "Existing Author", email = "author@example.org"}]
            license = {text = "Apache-2.0"}
            requires-python = ">=3.10"
            dependencies = ["requests"]

            [build-system]
            requires = ["hatchling"]
            build-backend = "hatchling.build"

            [tool.pdm]
            distribution = false
            """
        )
    )
    project_no_init.pyproject.reload()
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )

    result = pdm(["init"], input="\n\n\n\n\n\n\n\n\n", obj=project_no_init)

    assert result.exit_code == 0
    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        data = tomllib.load(fp)
    assert data["project"]["name"] == "existing-project"
    assert data["project"]["version"] == "2.1.0"
    assert data["project"]["description"] == "Existing description"
    assert data["project"]["authors"] == [{"name": "Existing Author", "email": "author@example.org"}]
    assert data["project"]["license"] == {"text": "Apache-2.0"}
    assert data["project"]["requires-python"] == ">=3.10"
    assert data["project"]["dependencies"] == ["requests"]
    assert data["build-system"] == {"requires": ["hatchling"], "build-backend": "hatchling.build"}
    assert data["tool"]["pdm"]["distribution"] is False


def test_init_updates_existing_pyproject_with_explicit_values(project_no_init, pdm, mocker):
    project_no_init.pyproject._path.write_text(
        dedent(
            """
            [project]
            name = "existing-project"
            version = "2.1.0"
            license = {text = "Apache-2.0"}
            requires-python = ">=3.10"
            dependencies = ["requests"]

            [build-system]
            requires = ["hatchling"]
            build-backend = "hatchling.build"

            [tool.pdm]
            distribution = false
            """
        )
    )
    project_no_init.pyproject.reload()
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )

    result = pdm(
        [
            "init",
            "--name",
            "updated-project",
            "--project-version",
            "3.0.0",
            "--license",
            "MIT",
            "--backend",
            "setuptools",
        ],
        input="\n\nUpdated description\n\n\n\n\n\n",
        obj=project_no_init,
    )

    assert result.exit_code == 0
    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        data = tomllib.load(fp)
    assert data["project"]["name"] == "updated-project"
    assert data["project"]["version"] == "3.0.0"
    assert data["project"]["description"] == "Updated description"
    assert data["project"]["license"] == {"text": "MIT"}
    assert data["project"]["dependencies"] == ["requests"]
    assert data["build-system"] == {"requires": ["setuptools>=61"], "build-backend": "setuptools.build_meta"}
    assert data["tool"]["pdm"]["distribution"] is True


def test_init_asks_whether_to_initialize_git(project_no_init, pdm, mocker):
    initialize_git = mocker.patch("pdm.cli.commands.init.Command.initialize_git")

    result = pdm(["init"], input="\ntest-project\n\n\n\n\n\n\nn\n", obj=project_no_init)

    assert result.exit_code == 0
    initialize_git.assert_not_called()


def test_init_no_git_does_not_initialize_git(project_no_init, pdm, mocker):
    initialize_git = mocker.spy(Command, "initialize_git")

    result = pdm(["init", "--no-git"], input="\ntest-project\n\n\n\n\n\n\n", obj=project_no_init)

    assert result.exit_code == 0
    initialize_git.assert_not_called()


def test_init_non_interactive_initializes_git_by_default(project_no_init, pdm, mocker):
    initialize_git = mocker.patch("pdm.cli.commands.init.Command.initialize_git")
    confirm = mocker.patch("pdm.cli.commands.init.termui.confirm")
    mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))

    result = pdm(["init", "-n"], obj=project_no_init)

    assert result.exit_code == 0
    confirm.assert_not_called()
    initialize_git.assert_called_once_with(project_no_init)


def test_new_command(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    with cd(project_no_init.root):
        pdm(
            ["new", "--name", "test-project", "--python", sys.executable, "myproject"],
            input="\n\n\n\n\n\n\n",
            strict=True,
        )
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("myproject/pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_adds_project_to_parent_workspace(project_no_init, pdm, mocker):
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    project_no_init.pyproject.settings["workspace"] = {"members": []}
    project_no_init.pyproject.write()

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "--no-git", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    do_use.assert_not_called()
    project_no_init.pyproject.reload()
    assert project_no_init.pyproject.settings["workspace"]["members"] == ["packages/foo"]
    assert not project_no_init.root.joinpath("packages/foo/.pdm-python").exists()


def test_init_creates_parent_workspace_table(project_no_init, pdm, mocker):
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    project_no_init.pyproject.write()

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "--no-git", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    do_use.assert_not_called()
    project_no_init.pyproject.reload()
    assert project_no_init.pyproject.settings["workspace"]["members"] == ["packages/foo"]
    assert not project_no_init.root.joinpath("packages/foo/.pdm-python").exists()


def test_init_workspace_member_does_not_persist_uv_workspace(project_no_init, pdm, mocker):
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    project_no_init.project_config["use_uv"] = True
    project_no_init.pyproject.write()

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "--no-git", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    do_use.assert_not_called()
    with project_no_init.root.joinpath("pyproject.toml").open("rb") as fp:
        data = tomllib.load(fp)
    assert data["tool"]["pdm"]["workspace"]["members"] == ["packages/foo"]
    assert "uv" not in data["tool"]


def test_init_workspace_member_skips_git_and_gitignore(project_no_init, pdm, mocker):
    initialize_git = mocker.patch("pdm.cli.commands.init.Command.initialize_git")
    project_no_init.pyproject.settings["workspace"] = {"members": []}
    project_no_init.pyproject.write()

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    initialize_git.assert_not_called()
    assert not project_no_init.root.joinpath("packages/foo/.git").exists()
    assert not project_no_init.root.joinpath("packages/foo/.gitignore").exists()


def test_init_workspace_member_keeps_existing_gitignore(project_no_init, pdm):
    project_no_init.pyproject.settings["workspace"] = {"members": []}
    project_no_init.pyproject.write()
    member_path = project_no_init.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    gitignore = member_path / ".gitignore"
    gitignore.write_text("local-cache/\n", encoding="utf-8")

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    assert gitignore.read_text(encoding="utf-8") == "local-cache/\n"


def test_init_workspace_member_reuses_root_requires_python(project_no_init, pdm, mocker):
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    project_no_init.pyproject.metadata["requires-python"] = ">=3.11"
    project_no_init.pyproject.settings["workspace"] = {"members": []}
    project_no_init.pyproject.write()

    with cd(project_no_init.root):
        result = pdm(
            ["init", "-n", "--no-git", "-p", "packages/foo", "--name", "foo"],
            cleanup=False,
        )

    assert result.exit_code == 0
    do_use.assert_not_called()
    with open(project_no_init.root.joinpath("packages/foo/pyproject.toml"), "rb") as fp:
        data = tomllib.load(fp)
    assert data["project"]["requires-python"] == ">=3.11"
    assert not project_no_init.root.joinpath("packages/foo/.pdm-python").exists()


def test_root_only_command_rejects_workspace_member(project_no_init, pdm):
    project_no_init.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project_no_init.pyproject.write()
    member_path = project_no_init.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\n',
        encoding="utf-8",
    )

    with cd(member_path):
        result = pdm(["install"], cleanup=False)

    assert result.exit_code == 1
    assert "can only be run from the workspace root" in result.stderr


def test_init_command_library(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    result = pdm(
        ["init"],
        input="\ntest-project\n\ny\nTest Project\n1\n\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Test Project",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f">={python_version}",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "build-system": {"build-backend": "setuptools.build_meta", "requires": ["setuptools>=61"]},
        "tool": {"pdm": {"distribution": True}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


@pytest.mark.parametrize(
    "backend_choice,merged_backend",
    [
        (0, {"build-backend": "pdm.backend", "requires": ["pdm-backend", "example"]}),
        (1, {"build-backend": "setuptools.build_meta", "requires": ["setuptools>=61"]}),
    ],
)
def test_init_template_build_system(tmp_path, project_no_init, pdm, mocker, backend_choice, merged_backend):
    template_with_backend = tmp_path / "backend-template"
    template_with_backend.mkdir()
    (template_with_backend / "pyproject.toml").write_text(
        dedent(
            """
            [project]
            dynamic = ["version"]
            name = "backend-template"

            [build-system]
            requires = ["pdm-backend", "example"]
            build-backend = "pdm.backend"
            """
        )
    )

    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    result = pdm(
        ["init", str(template_with_backend)],
        input=f"\ntest-project\n\ny\nTest Project\n{backend_choice}\n\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Test Project",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f">={python_version}",
            "dynamic": ["version"],
        },
        "build-system": merged_backend,
        "tool": {"pdm": {"distribution": True}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    result = pdm(["init", "-n"], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_auto_create_venv(project_no_init, pdm, mocker):
    mocker.patch("pdm.models.python.PythonInfo.get_venv", return_value=None)
    project_no_init.project_config["python.use_venv"] = True
    result = pdm(["init"], input="\ntest-project\n\ny\nTest Project\n1\n\n\n\n\n\n", obj=project_no_init)
    assert result.exit_code == 0
    assert project_no_init.python.executable.parent.parent == project_no_init.root / ".venv"
    assert ".pdm-python" in (project_no_init.root / ".gitignore").read_text()


def test_init_auto_create_venv_specify_python(project_no_init, pdm, mocker):
    mocker.patch("pdm.models.python.PythonInfo.get_venv", return_value=None)
    project_no_init.project_config["python.use_venv"] = True
    result = pdm(
        ["init", f"--python={PYTHON_VERSION}"],
        input="\n\n\n\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    assert project_no_init.python.executable.parent.parent == project_no_init.root / ".venv"


def test_init_with_backend_default_library(project_no_init, pdm):
    pdm(["init", "--backend", "flit-core"], input="\n\n\n\n\n\n\n\n\n", obj=project_no_init)
    assert project_no_init.backend.__class__.__name__ == "FlitBackend"


def test_init_with_backend_default_library_non_interactive(project_no_init, pdm):
    pdm(["init", "-n", "--backend", "flit-core"], obj=project_no_init)
    assert project_no_init.backend.__class__.__name__ == "FlitBackend"


def test_init_with_license_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    expected_license = "Proprietary"
    result = pdm(["init", "-n", "--license", expected_license], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": f"{expected_license}"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_with_project_version_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    expected_project_version = "2.0.42"
    result = pdm(["init", "-n", "--project-version", expected_project_version], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": f"{expected_project_version}",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data



import pytest

from pdm.cli import actions
from pdm.cli.commands.remove import Command
from pdm.cli.filters import GroupSelection
from pdm.models.specifiers import PySpecSet
from pdm.utils import cd


def make_workspace_member(project, core):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = ["requests"]\n',
        encoding="utf-8",
    )
    return core.create_project(member_path, global_config=project.global_config.config_file.as_posix())


def test_remove_command(project, pdm, mocker):
    do_remove = mocker.patch("pdm.cli.commands.remove.Command.do_remove")
    pdm(["remove", "demo"], obj=project)
    do_remove.assert_called_once()


@pytest.mark.usefixtures("working_set", "vcs")
def test_remove_editable_packages_while_keeping_normal(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "demo"], obj=project, strict=True)
    pdm(["add", "-d", "-e", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project, strict=True)
    pdm(["remove", "-d", "demo"], obj=project, strict=True)
    default_group = project.pyproject.metadata["dependencies"]
    dev_group = project.pyproject.dev_dependencies.get("dev")
    assert not dev_group
    assert len(default_group) == 1
    assert not project.get_locked_repository().candidates["demo"].req.editable


def test_remove_package(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "requests", "pytz"], obj=project, strict=True)
    pdm(["remove", *dev_option, "pytz"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert "pytz" not in locked_candidates
    assert "pytz" not in working_set


def test_remove_package_no_lock(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "requests", "pytz"], obj=project, strict=True)
    pdm(["remove", *dev_option, "--frozen-lockfile", "pytz"], obj=project, strict=True)
    assert "pytz" not in working_set
    project.lockfile.reload()
    locked_candidates = project.get_locked_repository().candidates
    assert "pytz" in locked_candidates


def test_remove_package_with_dry_run(project, working_set, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    result = pdm(["remove", "requests", "--dry-run"], obj=project, strict=True)
    project._lockfile = None
    locked_candidates = project.get_locked_repository().candidates
    assert "urllib3" in locked_candidates
    assert "urllib3" in working_set
    assert "- urllib3 1.22" in result.output


def test_remove_package_no_sync(project, working_set, pdm):
    pdm(["add", "requests", "pytz"], obj=project, strict=True)
    pdm(["remove", "pytz", "--no-sync"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert "pytz" not in locked_candidates
    assert "pytz" in working_set


@pytest.mark.usefixtures("working_set")
def test_remove_package_not_exist(project, pdm):
    pdm(["add", "requests", "pytz"], obj=project, strict=True)
    result = pdm(["remove", "django"], obj=project)
    assert result.exit_code == 1


def test_remove_package_exist_in_multi_groups(project, working_set, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    pdm(["add", "--dev", "urllib3"], obj=project, strict=True)
    pdm(["remove", "--dev", "urllib3"], obj=project, strict=True)
    assert "dependency-groups" not in project.pyproject._data
    assert "urllib3" in working_set
    assert "requests" in working_set


@pytest.mark.usefixtures("repository")
def test_remove_no_package(project, pdm):
    result = pdm(["remove"], obj=project)
    assert result.exit_code != 0


@pytest.mark.usefixtures("working_set")
def test_remove_package_wont_break_toml(project_no_init, pdm):
    project_no_init.pyproject._path.write_text(
        """
[project]
dependencies = [
    "requests",
    # this is a comment
]
"""
    )
    project_no_init.pyproject.reload()
    pdm(["remove", "requests"], obj=project_no_init, strict=True)
    assert project_no_init.pyproject.metadata["dependencies"] == []


@pytest.mark.usefixtures("working_set")
def test_remove_group_not_in_lockfile(project, pdm, mocker):
    pdm(["add", "requests"], obj=project, strict=True)
    project.add_dependencies(["pytz"], to_group="tz")
    assert project.lockfile.groups == ["default"]
    locker = mocker.patch.object(actions, "do_lock")
    pdm(["remove", "--group", "tz", "pytz"], obj=project, strict=True)
    assert "optional-dependencies" not in project.pyproject.metadata
    locker.assert_not_called()


def test_remove_workspace_member_updates_root_lockfile(project, core, repository):
    member_project = make_workspace_member(project, core)
    actions.do_lock(member_project)
    assert "requests" in member_project.get_locked_repository().candidates

    Command.do_remove(
        member_project,
        selection=GroupSelection(member_project),
        packages=["requests"],
        sync=False,
    )

    assert (project.root / "pdm.lock").exists()
    assert not (member_project.root / "pdm.lock").exists()
    locked_candidates = member_project.get_locked_repository().candidates
    assert "foo" in locked_candidates
    assert "requests" not in locked_candidates


def test_remove_subdirectory_path_dependency_keeps_workspace_member(project, repository):
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    project.pyproject.settings["workspace"] = {"members": ["packages/foo"]}
    project.pyproject.metadata["dependencies"] = ["foo @ file:///${PROJECT_ROOT}/packages/foo"]
    project.pyproject.write()

    Command.do_remove(
        project,
        selection=GroupSelection(project),
        packages=["foo"],
        sync=False,
    )

    project.pyproject.reload()
    assert project.pyproject.settings["workspace"]["members"] == ["packages/foo"]
    assert project.pyproject.metadata["dependencies"] == []
    assert "foo" in project.get_locked_repository().candidates


def test_remove_workspace_member_not_in_dependencies(project, repository):
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    project.pyproject.settings["workspace"] = {"members": ["packages/foo"]}
    project.pyproject.write()

    with cd(project.root):
        Command.do_remove(
            project,
            selection=GroupSelection(project),
            packages=["packages/foo"],
            sync=False,
        )

    project.pyproject.reload()
    assert project.pyproject.settings["workspace"]["members"] == []
    assert project.pyproject.metadata["dependencies"] == []
    assert "foo" not in project.get_locked_repository().candidates


@pytest.mark.usefixtures("working_set")
def test_remove_exclude_non_existing_dev_group_in_lockfile(project, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    project.add_dependencies(["pytz"], to_group="tz", dev=True)
    assert project.lockfile.groups == ["default"]
    result = pdm(["remove", "requests"], obj=project)
    assert result.exit_code == 0


@pytest.mark.usefixtures("working_set")
def test_remove_package_with_group_include(project, pdm):
    project.pyproject._data["dependency-groups"] = {
        "web": ["requests"],
        "serve": [{"include-group": "web"}, "django"],
    }
    project.pyproject.write()
    pdm(["lock"], obj=project, strict=True)
    pdm(["remove", "--no-sync", "-Gserve", "django"], obj=project, strict=True)
    assert "django" not in project.pyproject.dependency_groups["serve"]



import os
import platform
import re
import shutil
import subprocess
import sys
from pathlib import Path
from unittest.mock import ANY

import pytest
import shellingham

from pdm.cli.commands.venv import backends
from pdm.cli.commands.venv.utils import get_venv_prefix


@pytest.fixture(params=[True, False])
def with_pip(request):
    return request.param


@pytest.fixture()
def fake_create(monkeypatch):
    def fake_create(self, location, *args, prompt=None):
        bin_dir = "Scripts" if sys.platform == "win32" else "bin"
        suffix = ".exe" if sys.platform == "win32" else ""
        (location / bin_dir).mkdir(parents=True)
        (location / bin_dir / f"python{suffix}").touch()

    monkeypatch.setattr(backends.VirtualenvBackend, "perform_create", fake_create)
    monkeypatch.setattr(backends.VenvBackend, "perform_create", fake_create)
    monkeypatch.setattr(backends.CondaBackend, "perform_create", fake_create)


def test_core_registration_does_not_import_venv_runtime_helpers():
    code = """
import sys
from pdm.core import Core

Core()
lazy_modules = {
    "pdm.cli.commands.venv.backends",
    "pdm.cli.commands.venv.utils",
    "shellingham",
}
loaded = sorted(lazy_modules.intersection(sys.modules))
if loaded:
    print("\\n".join(loaded))
    raise SystemExit(1)
"""
    env = os.environ.copy()
    src = Path(__file__).resolve().parents[2] / "src"
    pythonpath = [str(src)]
    if env.get("PYTHONPATH"):
        pythonpath.append(env["PYTHONPATH"])
    env["PYTHONPATH"] = os.pathsep.join(pythonpath)
    result = subprocess.run([sys.executable, "-c", code], capture_output=True, env=env, text=True)

    assert result.returncode == 0, result.stdout or result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_create(pdm, project):
    project._saved_python = None
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    assert os.path.exists(venv_path)
    assert not project._saved_python


@pytest.mark.usefixtures("fake_create")
def test_venv_create_in_project(pdm, project):
    project.project_config["venv.in_project"] = True
    pdm(["venv", "create"], obj=project, strict=True)
    venv_path = project.root / ".venv"
    assert venv_path.exists()
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 1
    assert "is not empty" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_create_other_location(pdm, project):
    pdm(["venv", "-p", project.root.as_posix(), "create"], strict=True)
    venv_path = project.root / ".venv"
    assert venv_path.exists()
    result = pdm(["venv", "-p", project.root.as_posix(), "create"])
    assert result.exit_code == 1
    assert "is not empty" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_show_path(pdm, project):
    project.project_config["venv.in_project"] = True
    pdm(["venv", "create"], obj=project, strict=True)
    pdm(["venv", "create", "--name", "test"], obj=project, strict=True)
    result = pdm(["venv", "--path", "in-project"], obj=project, strict=True)
    assert result.output.strip() == str(project.root / ".venv")
    result = pdm(["venv", "--path", "test"], obj=project)
    assert result.exit_code == 0
    result = pdm(["venv", "--path", "foo"], obj=project)
    assert result.exit_code == 1


@pytest.mark.usefixtures("fake_create")
def test_venv_list(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)

    result = pdm(["venv", "list"], obj=project)
    assert result.exit_code == 0, result.stderr
    assert venv_path in result.output


@pytest.mark.usefixtures("fake_create")
def test_venv_remove(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    result = pdm(["venv", "remove", "non-exist"], obj=project)
    assert result.exit_code != 0

    result = pdm(["venv", "remove", "-y", key], obj=project)
    assert result.exit_code == 0, result.stderr

    assert not os.path.exists(venv_path)


@pytest.mark.usefixtures("fake_create")
def test_venv_recreate(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr

    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code != 0

    result = pdm(["venv", "create", "-f"], obj=project)
    assert result.exit_code == 0, result.stderr


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", return_value=("bash", None))
    result = pdm(["venv", "activate", key], obj=project)
    assert result.exit_code == 0, result.stderr
    backend = project.config["venv.backend"]

    if backend == "conda":
        assert result.output.startswith("conda activate")
    else:
        assert result.output.strip("'\"\n").endswith("activate")
        if platform.system() == "Windows":
            assert not result.output.startswith("source")
            assert not result.output.startswith("'")
        else:
            assert result.output.startswith("source")


@pytest.mark.usefixtures("venv_backends")
@pytest.mark.skipif(platform.system() == "Windows", reason="UNIX only")
def test_venv_activate_tcsh(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", return_value=("tcsh", None))
    result = pdm(["venv", "activate", key], obj=project)
    assert result.output.startswith("source") and result.output.strip("'\"\n").endswith("activate.csh")


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate_custom_prompt(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    creator.assert_called_once_with(
        None, [], False, False, prompt=project.project_config["venv.prompt"], with_pip=False
    )


def test_venv_activate_project_without_python(pdm, project):
    project._saved_python = None
    result = pdm(["venv", "activate"], obj=project)
    assert result.exit_code != 0
    assert "The project doesn't have a saved python.path" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_activate_error(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project, strict=True)

    result = pdm(["venv", "activate", "foo"], obj=project)
    assert result.exit_code != 0
    assert "No virtualenv with key" in result.stderr

    project._saved_python = os.path.abspath("fake/bin/python")
    result = pdm(["venv", "activate"], obj=project)
    assert result.exit_code != 0, result.output + result.stderr
    assert "Can't activate a non-venv Python" in result.stderr


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate_no_shell(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", side_effect=shellingham.ShellDetectionFailure())
    result = pdm(["venv", "activate", key], obj=project)
    assert result.exit_code == 0, result.stderr
    backend = project.config["venv.backend"]

    if backend == "conda":
        assert result.output.startswith("conda activate")
    else:
        assert result.output.strip("'\"\n").endswith("activate")
        if platform.system() == "Windows":
            assert not result.output.startswith("source")
            assert not result.output.startswith("'")
        else:
            assert result.output.startswith("source")


@pytest.mark.usefixtures("fake_create")
@pytest.mark.parametrize("keep_pypackages", [True, False])
def test_venv_auto_create(pdm, mocker, project, keep_pypackages):
    creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
    project._saved_python = None
    if keep_pypackages:
        project.root.joinpath("__pypackages__").mkdir(exist_ok=True)
    else:
        shutil.rmtree(project.root / "__pypackages__", ignore_errors=True)
    project.project_config["python.use_venv"] = True
    pdm(["install", "--no-self"], obj=project)
    if keep_pypackages:
        creator.assert_not_called()
    else:
        creator.assert_called_once()


@pytest.mark.usefixtures("fake_create")
def test_venv_purge(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "purge"], obj=project)
    assert result.exit_code == 0, result.stderr

    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge"], input="y", obj=project)
    assert result.exit_code == 0, result.stderr
    assert not os.path.exists(venv_path)


@pytest.mark.usefixtures("fake_create")
def test_venv_purge_force(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge", "-f"], obj=project)
    assert result.exit_code == 0, result.stderr
    assert not os.path.exists(venv_path)


user_options = [("none", True), ("0", False), ("all", False)]


@pytest.mark.usefixtures("venv_backends")
@pytest.mark.parametrize("user_choices, is_path_exists", user_options)
def test_venv_purge_interactive(pdm, user_choices, is_path_exists, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge", "-i"], input=user_choices, obj=project)
    assert result.exit_code == 0, result.stderr
    assert os.path.exists(venv_path) == is_path_exists


def test_virtualenv_backend_create(project, mocker, with_pip):
    backend = backends.VirtualenvBackend(project, None)
    assert backend.ident
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = [] if with_pip else ["--no-pip", "--no-setuptools", "--no-wheel"]
    mock_call.assert_called_once_with(
        [
            sys.executable,
            "-m",
            "virtualenv",
            str(location),
            "-p",
            str(backend._resolved_interpreter.executable),
            *pip_args,
        ],
        stdout=ANY,
    )


def test_venv_backend_create(project, mocker, with_pip):
    backend = backends.VenvBackend(project, None)
    assert backend.ident
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = [] if with_pip else ["--without-pip"]
    mock_call.assert_called_once_with(
        [
            str(backend._resolved_interpreter.executable),
            "-m",
            "venv",
            str(location),
            *pip_args,
        ],
        stdout=ANY,
    )


def test_conda_backend_create(project, mocker, with_pip):
    assert project.python
    backend = backends.CondaBackend(project, "3.9")
    assert backend.ident == "3.9"
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = ["pip"] if with_pip else []
    mock_call.assert_called_once_with(
        [
            "conda",
            "create",
            "--yes",
            "--prefix",
            str(location),
            "python=3.9",
            *pip_args,
        ],
        stdout=ANY,
    )

    backend = backends.CondaBackend(project, None)
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    assert backend.ident.startswith(python_version)
    location = backend.create()
    mock_call.assert_called_with(
        [
            "conda",
            "create",
            "--yes",
            "--prefix",
            str(location),
            f"python={python_version}",
        ],
        stdout=ANY,
    )



from __future__ import annotations

import argparse
import re
from os import PathLike
from pathlib import Path

from pdm.core import Core
from pdm.project.config import Config

ROOT = Path(__file__).resolve().parent.parent
REFERENCE_DIR = ROOT / "docs" / "reference"
TEMPLATE_DIR = ROOT / "tasks" / "doc_templates"
CLI_TEMPLATE = TEMPLATE_DIR / "cli.md.in"
CONFIGURATION_TEMPLATE = TEMPLATE_DIR / "configuration.md.in"
CLI_OUTPUT = REFERENCE_DIR / "cli.md"
CONFIGURATION_OUTPUT = REFERENCE_DIR / "configuration.md"
GENERATED_COMMENT = "<!-- Generated by tasks/render_reference_docs.py. -->"
MONOSPACED = ("pyproject.toml", "pdm.lock", ".pdm-python", ":pre", ":post", ":all")
ANSI_ESCAPE_RE = re.compile(r"\x1b\[[0-?]*[ -/]*[@-~]")


def clean_help(help_text: str | None) -> str:
    if not help_text:
        return ""
    help_text = strip_ansi_sequences(help_text)
    help_text = help_text.replace(r"\[", "[").replace(r"\]", "]")
    help_text = re.sub(r"__([\w\d_]+)__", r"`__\1__`", help_text)
    help_text = re.sub(r"env var: ([A-Z_]+)", r"env var: `\1`", help_text)
    help_text = re.sub(r"\[([^\]]+)\]", r"\\[\1\\]", help_text)
    for monospaced in MONOSPACED:
        help_text = re.sub(rf"\s(['\"]?{re.escape(monospaced)}['\"]?)", f" `{monospaced}`", help_text)
    return help_text


def render_parser(parser: argparse.ArgumentParser, title: str, heading_level: int = 2) -> str:
    result = [f"{'#' * heading_level} {title}\n"]
    if parser.description and title != "pdm":
        result.append("> " + clean_help(parser.description) + "\n")

    for group in sorted(parser._action_groups, key=lambda item: item.title.lower(), reverse=True):
        if not any(
            bool(action.option_strings or action.dest) or isinstance(action, argparse._SubParsersAction)
            for action in group._group_actions
        ):
            continue

        result.append(f"{group.title.title()}:\n")
        for action in group._group_actions:
            if isinstance(action, argparse._SubParsersAction):
                for name, subparser in action._name_parser_map.items():
                    result.append(render_parser(subparser, name, heading_level + 1))
                continue

            options = [f"`{option}`" for option in action.option_strings]
            line = f"- `{action.dest}`" if not options else f"- {', '.join(options)}"
            if action.metavar:
                line += f" `{action.metavar}`"
            line += f": {clean_help(action.help)}"

            if action.default and action.default != argparse.SUPPRESS:
                default = action.default
                if any(option.startswith("--no-") for option in action.option_strings) and default is True:
                    default = False
                line += f" (default: `{default}`)"

            result.append(line)
        result.append("")

    return "\n".join(result).rstrip()


def render_cli_reference() -> str:
    return render_parser(Core().parser, "pdm")


def render_configuration_table() -> str:
    rows = [
        "| Config Item | Description | Default Value | Available in Project | Env var |",
        "| --- | --- | --- | --- | --- |",
    ]

    for key, value in Config._config_map.items():
        default = f"`{format_default(value.default)}`" if value.should_show() else ""
        env_var = f"`{value.env_var}`" if value.env_var else ""
        available_in_project = "No" if value.global_only else "Yes"
        rows.append(f"| `{key}` | {value.description} | {default} | {available_in_project} | {env_var} |")

    rows.extend(
        [
            "| `pypi.<name>.url` | The URL of custom package source | `https://pypi.org/simple` | Yes | |",
            "| `pypi.<name>.username` | The username to access custom source | | Yes | |",
            "| `pypi.<name>.password` | The password to access custom source | | Yes | |",
            "| `pypi.<name>.type` | `index` or `find_links` | `index` | Yes | |",
            "| `pypi.<name>.verify_ssl` | Verify SSL certificate when query custom source | `True` | Yes | |",
            "| `repository.<name>.url` | The URL of custom package source | `https://pypi.org/simple` | Yes | |",
            "| `repository.<name>.username` | The username to access custom repository | | Yes | |",
            "| `repository.<name>.password` | The password to access custom repository | | Yes | |",
            (
                "| `repository.<name>.ca_certs` | Path to a PEM-encoded CA cert bundle "
                "(used for server cert verification) | The CA certificates from "
                "[certifi](https://pypi.org/project/certifi/) | Yes | |"
            ),
            "| `repository.<name>.verify_ssl` | Verify SSL certificate when uploading to repository | `True` | Yes | |",
        ]
    )
    return "\n".join(rows)


def format_default(value: object) -> str:
    if isinstance(value, PathLike):
        return normalize_home_path(Path(value))
    if isinstance(value, str):
        return normalize_home_path(strip_ansi_sequences(value))
    return str(value)


def normalize_home_path(value: str | Path) -> str:
    text = str(value)
    home = str(Path.home())
    if text.startswith(home):
        return text.replace(home, "~", 1)
    return text


def strip_ansi_sequences(value: str) -> str:
    return ANSI_ESCAPE_RE.sub("", value)


def render_template(path: Path, **replacements: str) -> str:
    content = path.read_text(encoding="utf-8")
    for key, value in replacements.items():
        content = content.replace(f"{{{{ {key} }}}}", value)
    return content


def write_if_changed(path: Path, content: str) -> None:
    normalized = content.rstrip() + "\n"
    if path.exists() and path.read_text(encoding="utf-8") == normalized:
        return
    path.write_text(normalized, encoding="utf-8")


def main() -> None:
    write_if_changed(
        CLI_OUTPUT,
        render_template(
            CLI_TEMPLATE,
            generated_comment=GENERATED_COMMENT,
            cli_reference=render_cli_reference(),
        ),
    )
    write_if_changed(
        CONFIGURATION_OUTPUT,
        render_template(
            CONFIGURATION_TEMPLATE,
            generated_comment=GENERATED_COMMENT,
            available_configurations=render_configuration_table(),
        ),
    )


if __name__ == "__main__":
    main()



import json
from unittest import mock

import pytest
from rich.box import ASCII


@mock.patch("pdm.termui.ROUNDED", ASCII)
@pytest.mark.usefixtures("working_set")
def test_outdated(project, pdm, index):
    pdm(["add", "requests"], obj=project, strict=True, cleanup=False)
    project.project_config["pypi.url"] = "https://my.pypi.org/simple"
    del project.pyproject.settings["source"]
    project.pyproject.write()
    index["/simple/requests/"] = b"""\
<!DOCTYPE html>
<html>
  <body>
    <h1>requests</h1>
    <a
      href="http://fixtures.test/artifacts/requests-2.20.0-py3-none-any.whl"
      data-requires-python=">=3.7"
    >
      requests-2.20.0-py3-none-any.whl
    </a>
  </body>
</html>

"""

    result = pdm(["outdated"], obj=project, strict=True, cleanup=False)
    assert "| requests | default | 2.19.1    | 2.19.1 | 2.20.0 |" in result.stdout

    result = pdm(["outdated", "re*"], obj=project, strict=True, cleanup=False)
    assert "| requests | default | 2.19.1    | 2.19.1 | 2.20.0 |" in result.stdout

    result = pdm(["outdated", "--json"], obj=project, strict=True, cleanup=False)
    json_output = json.loads(result.stdout)
    assert json_output == [
        {
            "package": "requests",
            "groups": ["default"],
            "installed_version": "2.19.1",
            "pinned_version": "2.19.1",
            "latest_version": "2.20.0",
        }
    ]



from __future__ import annotations

import shutil

import pytest
from unearth import Link

from pdm.exceptions import RequirementError
from pdm.models.candidates import Candidate
from pdm.models.requirements import Requirement, parse_requirement
from pdm.models.specifiers import PySpecSet
from pdm.utils import is_path_relative_to
from tests import FIXTURES


def to_lines(requirements: list[Requirement]) -> list[str]:
    return sorted([r.as_line() for r in requirements])


@pytest.mark.usefixtures("local_finder")
def test_parse_local_directory_metadata(project, is_editable):
    requirement_line = f"{(FIXTURES / 'projects/demo').as_posix()}"
    req = parse_requirement(requirement_line, is_editable)
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("vcs", "local_finder")
def test_parse_vcs_metadata(project, is_editable):
    requirement_line = "git+https://github.com/test-root/demo.git@master#egg=demo"
    req = parse_requirement(requirement_line, is_editable)
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"
    lockfile = candidate.as_lockfile_entry(project.root)
    assert lockfile["ref"] == "master"
    if is_editable:
        assert "revision" not in lockfile
    else:
        assert lockfile["revision"] == "1234567890abcdef"


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize(
    "requirement_line",
    [
        f"{(FIXTURES / 'artifacts/demo-0.0.1.tar.gz').as_posix()}",
        f"{(FIXTURES / 'artifacts/demo-0.0.1-py2.py3-none-any.whl').as_posix()}",
    ],
)
def test_parse_artifact_metadata(requirement_line, project):
    req = parse_requirement(requirement_line)
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("local_finder")
def test_parse_metadata_with_extras(project):
    req = parse_requirement(
        f"demo[tests,security] @ {(FIXTURES / 'artifacts/demo-0.0.1-py2.py3-none-any.whl').as_uri()}"
    )
    candidate = Candidate(req)
    prepared = candidate.prepare(project.environment)
    assert prepared.link.is_wheel
    assert to_lines(prepared.get_dependencies_from_metadata()) == [
        "pytest",
        'requests; python_version >= "3.6"',
    ]


@pytest.mark.usefixtures("local_finder")
def test_parse_remote_link_metadata(project):
    req = parse_requirement("http://fixtures.test/artifacts/demo-0.0.1-py2.py3-none-any.whl")
    candidate = Candidate(req)
    prepared = candidate.prepare(project.environment)
    assert prepared.link.is_wheel
    assert to_lines(prepared.get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("local_finder")
@pytest.mark.parametrize(
    "req_str",
    [
        "demo @ file:///${PROJECT_ROOT}/tests/fixtures/artifacts/demo-0.0.1-py2.py3-none-any.whl",
        "demo @ file:///${PROJECT_ROOT}/tests/fixtures/artifacts/demo-0.0.1.tar.gz",
        "demo @ file:///${PROJECT_ROOT}/tests/fixtures/projects/demo",
        "-e ./tests/fixtures/projects/demo",
        "-e file:///${PROJECT_ROOT}/tests/fixtures/projects/demo#egg=demo",
        "-e file:///${PROJECT_ROOT}/tests/fixtures/projects/demo#-with-hash#egg=demo",
    ],
)
def test_expand_project_root_in_url(req_str, core):
    project = core.create_project(FIXTURES.parent.parent)
    if req_str.startswith("-e "):
        req = parse_requirement(req_str[3:], True)
    else:
        req = parse_requirement(req_str)
    req.relocate(project.backend)
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    lockfile_entry = candidate.as_lockfile_entry(project.root)
    if "path" in lockfile_entry:
        assert lockfile_entry["path"].startswith("./")
    else:
        assert "${PROJECT_ROOT}" in lockfile_entry["url"]


@pytest.mark.usefixtures("local_finder")
def test_parse_project_file_on_build_error(project):
    req = parse_requirement(f"{(FIXTURES / 'projects/demo-failure').as_posix()}")
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [
        'chardet; os_name == "nt"',
        "idna",
    ]
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("local_finder")
def test_parse_project_file_on_build_error_with_extras(project):
    req = parse_requirement(f"{(FIXTURES / 'projects/demo-failure').as_posix()}")
    req.extras = ("security", "tests")
    candidate = Candidate(req)
    deps = to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata())
    assert 'requests; python_version >= "3.6"' in deps
    assert "pytest" in deps
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("local_finder")
def test_parse_project_file_on_build_error_no_dep(project):
    req = parse_requirement(f"{(FIXTURES / 'projects/demo-failure-no-dep').as_posix()}")
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == []
    assert candidate.name == "demo"
    assert candidate.version == "0.0.1"


@pytest.mark.usefixtures("local_finder")
def test_parse_poetry_project_metadata(project, is_editable):
    req = parse_requirement(f"{(FIXTURES / 'projects/poetry-demo').as_posix()}", is_editable)
    candidate = Candidate(req)
    requests_dep = "requests<3.0,>=2.6"
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == [requests_dep]
    assert candidate.name == "poetry-demo"
    assert candidate.version == "0.1.0"


@pytest.mark.usefixtures("local_finder")
def test_parse_flit_project_metadata(project, is_editable):
    req = parse_requirement(f"{(FIXTURES / 'projects/flit-demo').as_posix()}", is_editable)
    candidate = Candidate(req)
    deps = to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata())
    requests_dep = "requests>=2.6"
    assert requests_dep in deps
    assert 'configparser; python_version == "2.7"' in deps
    assert candidate.name == "pyflit"
    assert candidate.version == "0.1.0"


@pytest.mark.usefixtures("vcs", "local_finder")
def test_vcs_candidate_in_subdirectory(project, is_editable):
    line = "git+https://github.com/test-root/demo-parent-package.git@master#egg=package-a&subdirectory=package-a"
    req = parse_requirement(line, is_editable)
    candidate = Candidate(req)
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == ["flask"]
    assert candidate.version == "0.1.0"

    line = "git+https://github.com/test-root/demo-parent-package.git@master#egg=package-b&subdirectory=package-b"
    req = parse_requirement(line, is_editable)
    candidate = Candidate(req)
    expected_deps = ["django"]
    assert to_lines(candidate.prepare(project.environment).get_dependencies_from_metadata()) == expected_deps
    tail = "+editable" if is_editable else ""
    assert candidate.version == f"0.1.0{tail}"


@pytest.mark.usefixtures("local_finder")
def test_sdist_candidate_with_wheel_cache(project, mocker):
    file_link = Link((FIXTURES / "artifacts/demo-0.0.1.tar.gz").as_uri())
    built_path = FIXTURES / "artifacts/demo-0.0.1-py2.py3-none-any.whl"
    wheel_cache = project.make_wheel_cache()
    cache_path = wheel_cache.get_path_for_link(file_link, project.environment.spec)
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    shutil.copy2(built_path, cache_path)
    req = parse_requirement(file_link.url)
    downloader = mocker.patch("unearth.finder.unpack_link")
    prepared = Candidate(req).prepare(project.environment)
    prepared.metadata
    downloader.assert_not_called()
    assert prepared._cached.name == built_path.name

    prepared._cached = None
    builder = mocker.patch("pdm.builders.WheelBuilder.build")
    wheel = prepared.build()
    builder.assert_not_called()
    assert wheel.name == built_path.name


@pytest.mark.usefixtures("vcs", "local_finder")
def test_cache_vcs_immutable_revision(project):
    req = parse_requirement("git+https://github.com/test-root/demo.git@master#egg=demo")
    candidate = Candidate(req)
    candidate.prepare(project.environment).build()
    assert not is_path_relative_to(candidate.prepared._get_build_cache(), project.cache_dir)
    assert candidate.get_revision() == "1234567890abcdef"

    req = parse_requirement("git+https://github.com/test-root/demo.git@1234567890abcdef#egg=demo")
    candidate = Candidate(req)
    candidate.prepare(project.environment).build()
    assert is_path_relative_to(candidate.prepared._get_build_cache(), project.cache_dir)
    assert candidate.get_revision() == "1234567890abcdef"

    # test the revision can be got correctly after cached
    prepared = Candidate(req).prepare(project.environment)
    assert not prepared._source_dir
    assert prepared.revision == "1234567890abcdef"


@pytest.mark.usefixtures("local_finder")
def test_cache_egg_info_sdist(project):
    req = parse_requirement("demo @ http://fixtures.test/artifacts/demo-0.0.1.tar.gz")
    candidate = Candidate(req)
    candidate.prepare(project.environment).build()
    assert is_path_relative_to(candidate.prepared._get_build_cache(), project.cache_dir)


def test_invalidate_incompatible_wheel_link(project):
    project.project_config["pypi.url"] = "https://my.pypi.org/simple"
    project.environment.python_requires = PySpecSet(">=3.6")
    project.environment.__dict__["spec"] = project.environment.spec.replace(platform="linux")
    req = parse_requirement("demo")
    prepared = Candidate(
        req,
        name="demo",
        version="0.0.1",
        link=Link("http://fixtures.test/artifacts/demo-0.0.1-cp36-cp36m-win_amd64.whl"),
    ).prepare(project.environment)
    prepared._obtain(True)
    assert prepared._cached.name == prepared.link.filename == "demo-0.0.1-cp36-cp36m-win_amd64.whl"

    prepared._obtain(False)
    assert prepared._cached.name == prepared.link.filename == "demo-0.0.1-py2.py3-none-any.whl"


def test_legacy_pep345_tag_link(project):
    project.project_config["pypi.url"] = "https://my.pypi.org/simple"
    req = parse_requirement("pep345-legacy")
    repo = project.get_repository()
    _ = next(iter(repo.find_candidates(req)))


@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_ignore_invalid_py_version(project):
    project.project_config["pypi.url"] = "https://my.pypi.org/simple"
    req = parse_requirement("wheel")
    repo = project.get_repository()
    _ = next(iter(repo.find_candidates(req)))


def test_find_candidates_from_find_links(project):
    project.pyproject.settings["source"] = [
        {"name": "pypi", "url": "http://fixtures.test/index/demo.html", "verify_ssl": False, "type": "find_links"}
    ]
    project.pyproject.write(False)
    repo = project.get_repository()
    candidates = list(repo.find_candidates(parse_requirement("demo")))
    assert len(candidates) == 1
    assert candidates[0].link.filename == "demo-0.0.1-py2.py3-none-any.whl"


def test_parse_metadata_from_pep621(project, mocker):
    builder = mocker.patch("pdm.builders.wheel.WheelBuilder.build")
    req = parse_requirement(f"test-hatch @ {(FIXTURES / 'projects/test-hatch-static').as_uri()}")
    candidate = Candidate(req)
    distribution = candidate.prepare(project.environment).metadata
    assert sorted(distribution.requires) == ["click", "requests"]
    assert distribution.metadata["Summary"] == "Test hatch project"
    builder.assert_not_called()


def test_parse_metadata_with_dynamic_fields(project, local_finder):
    req = parse_requirement(f"demo-package @ {(FIXTURES / 'projects/demo-src-package').as_uri()}")
    candidate = Candidate(req)
    metadata = candidate.prepare(project.environment).metadata
    assert not metadata.requires
    assert metadata.version == "0.1.0"


def test_get_metadata_for_non_existing_path(project):
    req = parse_requirement("file:///${PROJECT_ROOT}/non-existing-path")
    candidate = Candidate(req)
    with pytest.raises(RequirementError, match=r"The local path '.+' does not exist"):
        candidate.prepare(project.environment).metadata



from pathlib import Path

import pycomplete

from pdm.core import Core

COMPLETIONS = Path(__file__).parent.parent / "src/pdm/cli/completions"


def main():
    core = Core()
    core.init_parser()

    completer = pycomplete.Completer(core.parser, ["pdm"])
    for shell in ("bash", "fish"):
        COMPLETIONS.joinpath(f"pdm.{shell}").write_text(completer.render(shell))


if __name__ == "__main__":
    main()



import io
import os
import shlex
import zipfile
from pathlib import Path

import pytest

from pdm.environments.local import PythonLocalEnvironment, _get_shebang_path, _is_console_script, _replace_shebang


@pytest.mark.parametrize(
    "executable,is_launcher,expected",
    [
        ("/usr/bin/python", True, b"/usr/bin/python"),
        ("/usr/bin/python", False, b"/usr/bin/python"),
        ("/a path/with space/python", False, shlex.quote("/a path/with space/python").encode("utf-8")),
    ],
)
def test_get_shebang_path(executable, is_launcher, expected):
    assert _get_shebang_path(executable, is_launcher) == expected


def test_is_console_script_unix_and_binary():
    # Non-Windows path: startswith #! => True
    assert _is_console_script(b"#!/usr/bin/env python\nprint('x')\n") is True
    # Undecodable bytes => False
    assert _is_console_script(b"\xff\xfe\xfd") is False


@pytest.mark.skipif(os.name == "nt", reason="Regex branch differs on Windows")
@pytest.mark.parametrize(
    "content,new_exec,expected_prefix",
    [
        (b"#!/usr/bin/python\nprint('hi')\n", b"/new/python", b"#!/new/python\n"),
        (
            b"#!/bin/sh\n'''exec' '/old path/python' \"$0\" \"$@\"\n' '''\nprint('x')\n",
            b"/new/python",
            b"#!/bin/sh\n'''exec' /new/python \"$0\"",
        ),
    ],
)
def test_replace_shebang_unix(tmp_path: Path, content: bytes, new_exec: bytes, expected_prefix: bytes):
    f = tmp_path / "script"
    f.write_bytes(content)
    _replace_shebang(f, new_exec)
    data = f.read_bytes()
    assert data.startswith(expected_prefix)


def test_update_shebangs_changes_scripts_header(project):
    env = PythonLocalEnvironment(project)
    # Create a fake script under the environment's scripts dir
    scripts = Path(env.get_paths()["scripts"])  # ensure exists
    script = scripts / "demo"
    script.write_text("#!/usr/bin/python\nprint('ok')\n", encoding="utf-8")

    new_path = "/opt/python/bin/python"
    # Exercise update_shebangs
    env.update_shebangs(new_path)

    text = script.read_text(encoding="utf-8")
    assert text.splitlines()[0] == f"#!{new_path}"


def test_update_shebangs_ignores_non_target_files_and_dirs(project):
    env = PythonLocalEnvironment(project)
    scripts = Path(env.get_paths()["scripts"])  # ensure exists
    # A non-target extension file
    other = scripts / "notascript.sh"
    other.write_text("#!/usr/bin/python\nprint('no change')\n", encoding="utf-8")
    # A directory
    d = scripts / "adir"
    d.mkdir(exist_ok=True)

    before_other = other.read_text(encoding="utf-8")
    env.update_shebangs("/opt/python/bin/python")
    after_other = other.read_text(encoding="utf-8")
    # Should be unchanged
    assert before_other == after_other


def test_replace_shebang_early_return_when_not_console(tmp_path: Path):
    f = tmp_path / "no_shebang"
    original = b"print('hello')\n"
    f.write_bytes(original)
    _replace_shebang(f, b"/new/python")
    assert f.read_bytes() == original


def _zip_with_main_bytes() -> bytes:
    buf = io.BytesIO()
    with zipfile.ZipFile(buf, mode="w") as zf:
        zf.writestr("__main__.py", "print('ok')\n")
    return buf.getvalue()


def test_is_console_script_windows_zip(monkeypatch):
    monkeypatch.setattr(os, "name", "nt")
    data = _zip_with_main_bytes()
    assert _is_console_script(data) is True
    # Not a valid zip => False (falls back to text detection which returns False)
    assert _is_console_script(b"not a zip") is False


def test_is_console_script_windows_text_shebang(monkeypatch):
    # On Windows, non-zip scripts with shebang should still be recognized
    monkeypatch.setattr(os, "name", "nt")
    assert _is_console_script(b"#!/usr/bin/env python\r\nprint('x')\r\n") is True


def test_replace_shebang_windows_zip_no_change(monkeypatch, tmp_path: Path):
    monkeypatch.setattr(os, "name", "nt")
    f = tmp_path / "script.exe"
    data = _zip_with_main_bytes()
    f.write_bytes(data)
    _replace_shebang(f, b"C:/Python/python.exe")
    # Should remain unchanged since regex won't match zip content
    assert f.read_bytes() == data



import shutil

import pytest
from unearth import Link

from pdm.cli.commands.add import Command
from pdm.cli.filters import GroupSelection
from pdm.models.markers import EnvSpec
from pdm.models.specifiers import PySpecSet
from pdm.pytest import Distribution
from pdm.utils import cd
from tests import FIXTURES


def make_workspace_member(project, core):
    project.pyproject.settings["workspace"] = {"members": ["packages/*"]}
    project.pyproject.write()
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )
    return core.create_project(member_path, global_config=project.global_config.config_file.as_posix())


def test_add_package(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "requests"], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]

    assert group[0] == "requests>=2.19.1"
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_package_no_lock(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "--frozen-lockfile", "-v", "requests"], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]

    assert group[0] == "requests>=2.19.1"
    assert not project.lockfile.exists()
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_command(project, pdm, mocker):
    do_add = mocker.patch("pdm.cli.commands.add.Command.do_add")
    pdm(["add", "requests"], obj=project)
    do_add.assert_called_once()


def test_add_workspace_member_updates_root_lockfile(project, core, repository):
    member_project = make_workspace_member(project, core)

    Command.do_add(
        member_project,
        selection=GroupSelection(member_project),
        packages=["requests"],
        sync=False,
    )

    assert (project.root / "pdm.lock").exists()
    assert not (member_project.root / "pdm.lock").exists()
    locked_candidates = member_project.get_locked_repository().candidates
    assert "foo" in locked_candidates
    assert "requests" in locked_candidates


def test_add_subdirectory_path_updates_workspace(project, repository):
    member_path = project.root / "packages" / "foo"
    member_path.mkdir(parents=True)
    member_path.joinpath("pyproject.toml").write_text(
        '[project]\nname = "foo"\nversion = "0.1.0"\ndependencies = []\n',
        encoding="utf-8",
    )

    with cd(project.root):
        Command.do_add(
            project,
            selection=GroupSelection(project),
            packages=["packages/foo"],
            sync=False,
        )

    project.pyproject.reload()
    assert project.pyproject.settings["workspace"]["members"] == ["packages/foo"]
    assert project.pyproject.metadata["dependencies"] == ["foo @ file:///${PROJECT_ROOT}/packages/foo"]
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["foo"].req.editable
    assert locked_candidates["foo"].req.str_path == "./packages/foo"


def test_add_package_to_custom_group(project, working_set, pdm):
    pdm(["add", "requests", "--group", "test"], obj=project, strict=True)

    assert "requests" in project.pyproject.metadata["optional-dependencies"]["test"][0]
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_package_to_custom_dev_group(project, working_set, pdm):
    pdm(["add", "requests", "--group", "test", "--dev"], obj=project, strict=True)

    dependencies = project.pyproject.dependency_groups["test"]
    assert "requests" in dependencies[0]
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


@pytest.mark.usefixtures("vcs")
def test_add_editable_package(project, working_set, pdm):
    # Ensure that correct python version is used.
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--dev", "demo"], obj=project, strict=True)
    pdm(["add", "-de", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project, strict=True)

    group = project.pyproject.dev_dependencies["dev"]
    assert group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert not project.pyproject.dependency_groups
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["demo"].prepare(project.environment).revision == "1234567890abcdef"
    assert working_set["demo"].link_file
    assert locked_candidates["idna"].version == "2.7"
    assert "idna" in working_set

    pdm(["sync", "--no-editable"], obj=project, strict=True)
    assert not working_set["demo"].link_file


@pytest.mark.usefixtures("vcs", "working_set")
def test_add_editable_package_to_metadata_forbidden(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    result = pdm(["add", "-v", "-e", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project)
    assert "PdmUsageError" in result.stderr
    result = pdm(["add", "-v", "-Gtest", "-e", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project)
    assert "PdmUsageError" in result.stderr


@pytest.mark.usefixtures("working_set", "vcs")
def test_non_editable_override_editable(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "--dev", "-e", url], obj=project, strict=True)
    pdm(["add", "--dev", url], obj=project, strict=True)
    assert not project.get_dependencies("dev")[0].editable


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_editable_normal_dev_dependency(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "--dev", "-e", url], obj=project, strict=True)
    pdm(["add", "-d", "urllib3"], obj=project, strict=True)
    pdm(["add", "-d", "idna"], obj=project, strict=True)
    dev_group = project.pyproject.settings["dev-dependencies"]["dev"]
    pep735_group = project.pyproject.dependency_groups["dev"]
    assert dev_group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert pep735_group == ["urllib3>=1.22", "idna>=2.7"]


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_dev_dependency_with_existing_editables_group(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "-dG", "editables", "-e", url], obj=project, strict=True)
    pdm(["add", "-d", "urllib3"], obj=project, strict=True)
    pdm(["add", "-dG", "named", "idna"], obj=project, strict=True)
    assert "editables" in project.pyproject.settings["dev-dependencies"]
    assert "dev" in project.pyproject.dependency_groups
    assert "named" in project.pyproject.dependency_groups
    editables_group = project.pyproject.settings["dev-dependencies"]["editables"]
    pep735_group = project.pyproject.dependency_groups["dev"]
    pep735_named_group = project.pyproject.dependency_groups["named"]
    assert editables_group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert "editables" not in project.pyproject.dependency_groups
    assert pep735_group == ["urllib3>=1.22"]
    assert pep735_named_group == ["idna>=2.7"]


@pytest.mark.usefixtures("working_set")
def test_add_remote_package_url(project, dev_option, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "http://fixtures.test/artifacts/demo-0.0.1-py2.py3-none-any.whl"
    pdm(["add", *dev_option, url], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]
    assert group[0] == f"demo @ {url}"


def test_add_no_install(project, working_set, pdm):
    pdm(["add", "--no-sync", "requests"], obj=project, strict=True)
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package not in working_set


def test_add_no_sync_frozen_lockfile_skips_resolution(project, pdm, mocker):
    do_lock = mocker.patch("pdm.cli.actions.do_lock")
    pdm(["add", "--frozen-lockfile", "--no-sync", "requests"], obj=project, strict=True)

    do_lock.assert_not_called()
    assert project.pyproject.metadata["dependencies"][0] == "requests"
    assert not project.lockfile.exists()


@pytest.mark.usefixtures("repository")
def test_add_package_save_exact(project, pdm):
    pdm(["add", "--save-exact", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests==2.19.1"


@pytest.mark.usefixtures("repository")
def test_add_package_save_wildcard(project, pdm):
    pdm(["add", "--save-wildcard", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests"


@pytest.mark.usefixtures("repository")
def test_add_package_save_minimum(project, pdm):
    pdm(["add", "--save-minimum", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests>=2.19.1"


def test_add_package_update_reuse(project, repository, pdm):
    pdm(["add", "--no-sync", "--save-wildcard", "requests", "pytz"], obj=project, strict=True)

    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.19.1"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"

    repository.add_candidate("pytz", "2019.6")
    repository.add_candidate("chardet", "3.0.5")
    repository.add_candidate("requests", "2.20.0")
    repository.add_dependencies(
        "requests",
        "2.20.0",
        [
            "certifi>=2017.4.17",
            "chardet<3.1.0,>=3.0.2",
            "idna<2.8,>=2.5",
            "urllib3<1.24,>=1.21.1",
        ],
    )
    pdm(["add", "--no-sync", "--save-wildcard", "requests"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.20.0"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"


def test_add_package_update_eager(project, repository, pdm):
    pdm(["add", "--no-sync", "--save-wildcard", "requests", "pytz"], obj=project, strict=True)

    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.19.1"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"

    repository.add_candidate("pytz", "2019.6")
    repository.add_candidate("chardet", "3.0.5")
    repository.add_candidate("requests", "2.20.0")
    repository.add_dependencies(
        "requests",
        "2.20.0",
        [
            "certifi>=2017.4.17",
            "chardet<3.1.0,>=3.0.2",
            "idna<2.8,>=2.5",
            "urllib3<1.24,>=1.21.1",
        ],
    )
    pdm(["add", "--no-sync", "--save-wildcard", "--update-eager", "requests"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.20.0"
    assert locked_candidates["chardet"].version == "3.0.5"
    assert locked_candidates["pytz"].version == "2019.3"


def test_add_package_with_mismatch_marker(project, working_set, pdm):
    env = project.environment
    env.__dict__["spec"] = EnvSpec.from_spec("==3.11", "macos", "cpython")
    pdm(["add", "requests", "pytz; platform_system!='Darwin'"], obj=project, strict=True)
    assert "pytz" not in working_set


def test_add_dependency_from_multiple_parents(project, working_set, pdm):
    env = project.environment
    env.__dict__["spec"] = EnvSpec.from_spec("==3.11", "macos", "cpython")
    pdm(["add", "requests", "chardet; platform_system!='Darwin'"], obj=project, strict=True)
    assert "chardet" in working_set


def test_add_packages_without_self(project, working_set, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--no-self", "requests"], obj=project, strict=True)
    assert project.name not in working_set


@pytest.mark.usefixtures("working_set")
def test_add_package_unconstrained_rewrite_specifier(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--no-self", "django"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["django"].version == "2.2.9"
    assert project.pyproject.metadata["dependencies"][0] == "django>=2.2.9"

    pdm(["add", "--no-self", "--unconstrained", "django-toolbar"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["django"].version == "1.11.8"
    assert project.pyproject.metadata["dependencies"][0] == "django>=1.11.8"


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_cached_vcs_requirement(project, mocker, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git@1234567890abcdef#egg=demo"
    built_path = FIXTURES / "artifacts/demo-0.0.1-py2.py3-none-any.whl"
    wheel_cache = project.make_wheel_cache()
    cache_path = wheel_cache.get_path_for_link(Link(url), project.environment.spec)
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    shutil.copy2(built_path, cache_path)
    downloader = mocker.patch("unearth.finder.unpack_link")
    builder = mocker.patch("pdm.builders.WheelBuilder.build")
    pdm(["add", "--no-self", url], obj=project, strict=True)
    lockfile_entry = next(p for p in project.lockfile["package"] if p["name"] == "demo")
    assert lockfile_entry["revision"] == "1234567890abcdef"
    downloader.assert_not_called()
    builder.assert_not_called()


@pytest.mark.usefixtures("repository")
def test_add_with_dry_run(project, pdm):
    result = pdm(["add", "--dry-run", "requests"], obj=project, strict=True)
    project.pyproject.reload()
    assert not project.get_dependencies()
    assert "requests 2.19.1" in result.stdout
    assert "urllib3 1.22" in result.stdout


def test_add_with_prerelease(project, working_set, pdm):
    pdm(["add", "--prerelease", "--save-compatible", "urllib3"], obj=project, strict=True)
    assert working_set["urllib3"].version == "1.23b0"
    assert project.pyproject.metadata["dependencies"][0] == "urllib3<2,>=1.23b0"


def test_add_editable_package_with_extras(project, working_set, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    dep_path = FIXTURES.joinpath("projects/demo")
    pdm(["add", "-dGdev", "-e", f"{dep_path.as_posix()}[security]"], obj=project, strict=True)
    assert f"-e {dep_path.as_uri()}#egg=demo[security]" in project.use_pyproject_dependencies("dev", True)[0]
    assert "demo" in working_set
    assert "requests" in working_set
    assert "urllib3" in working_set


@pytest.mark.usefixtures("repository")
def test_add_dependency_with_extras(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "requests[security]", "--no-sync"], obj=project, strict=True)
    locked_repo = project.get_locked_repository()
    extra_package = next(p for p in locked_repo.packages.values() if p.candidate.identify() == "requests[security]")
    assert extra_package.dependencies == ["pyOpenSSL>=0.14", "requests==2.19.1"]
    # test adding new package won't add duplicate dependencies
    pdm(["add", "pytz", "--no-sync"], obj=project, strict=True)
    locked_repo = project.get_locked_repository()
    extra_package = next(p for p in locked_repo.packages.values() if p.candidate.identify() == "requests[security]")
    assert extra_package.dependencies == ["pyOpenSSL>=0.14", "requests==2.19.1"]


def test_add_package_with_local_version(project, repository, working_set, pdm):
    repository.add_candidate("foo", "1.0a0+local")
    pdm(["add", "-v", "foo"], obj=project, strict=True)
    assert working_set["foo"].version == "1.0a0+local"
    dependencies, _ = project.use_pyproject_dependencies("default")
    assert dependencies[0] == "foo>=1.0a0"


def test_add_group_to_lockfile(project, working_set, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    assert project.lockfile.groups == ["default"]
    pdm(["add", "--group", "tz", "pytz"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "tz"]
    assert "pytz" in working_set


def test_add_group_to_lockfile_without_package(project, working_set, pdm):
    project.add_dependencies(["requests"])
    project.add_dependencies(["pytz"], to_group="tz")
    pdm(["install"], obj=project, strict=True)
    assert "pytz" not in working_set
    assert project.lockfile.groups == ["default"]
    pdm(["add", "--group", "tz"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "tz"]
    assert "pytz" in working_set


def test_add_update_reuse_installed(project, working_set, repository, pdm):
    working_set["foo"] = Distribution("foo", "1.0.0")
    repository.add_candidate("foo", "1.0.0")
    repository.add_candidate("foo", "1.1.0")
    pdm(["add", "--update-reuse-installed", "foo"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["foo"].version == "1.0.0"


def test_add_update_reuse_installed_config(project, working_set, repository, pdm):
    working_set["foo"] = Distribution("foo", "1.0.0")
    repository.add_candidate("foo", "1.0.0")
    repository.add_candidate("foo", "1.1.0")
    project.project_config["strategy.update"] = "reuse-installed"
    pdm(["add", "foo"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["foo"].version == "1.0.0"


def test_add_disable_cache(project, pdm, working_set):
    cache_dir = project.cache_dir
    pdm(["--no-cache", "add", "requests"], obj=project, strict=True)
    assert "requests" in working_set

    files = [file for file in cache_dir.rglob("*") if file.is_file()]
    assert not files


@pytest.mark.usefixtures("working_set")
def test_add_dependency_with_direct_minimal_versions(project, pdm, repository):
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    repository.add_candidate("pytz", "2019.6")
    pdm(["add", "django"], obj=project, strict=True)
    all_candidates = project.get_locked_repository().candidates
    assert "django>=1.11.8" in project.pyproject.metadata["dependencies"]
    assert all_candidates["django"].version == "1.11.8"
    assert all_candidates["pytz"].version == "2019.6"


def test_add_group_with_normalized_name(project, pdm, working_set):
    project.pyproject.dependency_groups.update({"foo_bar": ["requests"]})
    project.pyproject.write()
    pdm(["lock"], obj=project, strict=True)
    assert "foo-bar" in project.lockfile.groups
    pdm(["sync", "-G", "foo.bar"], obj=project, strict=True)
    assert "requests" in working_set
    result = pdm(["add", "-G", "foo-bar", "pytz"], obj=project)
    assert result.exit_code != 0
    assert "Group foo-bar already exists in another non-normalized form" in result.stderr


@pytest.mark.usefixtures("working_set")
def test_add_to_dependency_group_with_include(project, pdm):
    from pdm.formats.base import make_array

    project.pyproject.dependency_groups.update({"tz": ["pytz"], "web": make_array([{"include-group": "tz"}])})
    project.pyproject.write()
    pdm(["add", "-Gweb", "requests"], obj=project, strict=True)
    assert project.pyproject.dependency_groups["web"] == [{"include-group": "tz"}, "requests>=2.19.1"]



import pytest

from pdm.exceptions import PdmUsageError
from pdm.project.config import Config
from pdm.utils import cd


def test_config_command(project, pdm):
    result = pdm(["config"], obj=project)
    assert result.exit_code == 0
    assert "python.use_pyenv = True" in result.output

    result = pdm(["config", "-v"], obj=project)
    assert result.exit_code == 0
    assert "Use the pyenv interpreter" in result.output


def test_config_get_command(project, pdm):
    result = pdm(["config", "python.use_pyenv"], obj=project)
    assert result.exit_code == 0
    assert result.output.strip() == "True"

    result = pdm(["config", "foo.bar"], obj=project)
    assert result.exit_code != 0


def test_config_set_command(project, pdm):
    result = pdm(["config", "python.use_pyenv", "false"], obj=project)
    assert result.exit_code == 0
    result = pdm(["config", "python.use_pyenv"], obj=project)
    assert result.output.strip() == "False"

    result = pdm(["config", "foo.bar"], obj=project)
    assert result.exit_code != 0

    result = pdm(["config", "-l", "cache_dir", "/path/to/bar"], obj=project)
    assert result.exit_code != 0


def test_config_del_command(project, pdm):
    result = pdm(["config", "-l", "python.use_pyenv", "false"], obj=project)
    assert result.exit_code == 0

    result = pdm(["config", "python.use_pyenv"], obj=project)
    assert result.output.strip() == "False"

    result = pdm(["config", "-ld", "python.use_pyenv"], obj=project)
    assert result.exit_code == 0

    result = pdm(["config", "python.use_pyenv"], obj=project)
    assert result.output.strip() == "True"


def test_config_env_var_shadowing(project, pdm, monkeypatch):
    monkeypatch.setenv("PDM_PYPI_URL", "https://example.org/simple")
    result = pdm(["config", "pypi.url"], obj=project)
    assert result.output.strip() == "https://example.org/simple"

    result = pdm(["config", "pypi.url", "https://test.pypi.org/pypi"], obj=project)
    assert "config is shadowed by env var 'PDM_PYPI_URL'" in result.stderr
    result = pdm(["config", "pypi.url"], obj=project)
    assert result.output.strip() == "https://example.org/simple"

    monkeypatch.delenv("PDM_PYPI_URL")
    result = pdm(["config", "pypi.url"], obj=project)
    assert result.output.strip() == "https://test.pypi.org/pypi"


def test_config_project_global_precedence(project, pdm):
    pdm(["config", "python.use_pyenv", "true"], obj=project)
    pdm(["config", "-l", "python.use_pyenv", "false"], obj=project)

    result = pdm(["config", "python.use_pyenv"], obj=project)
    assert result.output.strip() == "False"


def test_specify_config_file(tmp_path, pdm, monkeypatch):
    tmp_path.joinpath("global_config.toml").write_text("strategy.resolve_max_rounds = 1000\n")
    with cd(tmp_path):
        result = pdm(["-c", "global_config.toml", "config", "strategy.resolve_max_rounds"])
        assert result.exit_code == 0
        assert result.output.strip() == "1000"

        monkeypatch.setenv("PDM_CONFIG_FILE", "global_config.toml")
        result = pdm(["config", "strategy.resolve_max_rounds"])
        assert result.exit_code == 0
        assert result.output.strip() == "1000"


def test_default_repository_setting(project):
    repository = project.global_config.get_repository_config("pypi", "repository")
    assert repository.url == "https://upload.pypi.org/legacy/"
    assert repository.username is None
    assert repository.password is None

    repository = project.global_config.get_repository_config("testpypi", "repository")
    assert repository.url == "https://test.pypi.org/legacy/"

    repository = project.global_config.get_repository_config("nonexist", "repository")
    assert repository is None


def test_repository_config_key_short(project):
    with pytest.raises(PdmUsageError):
        project.global_config["repository.test"] = {"url": "https://example.org/simple"}

    with pytest.raises(PdmUsageError):
        project.global_config["repository"] = "123"

    with pytest.raises(PdmUsageError):
        del project.global_config["repository"]


def test_repository_overwrite_default(project):
    project.global_config["repository.pypi.username"] = "foo"
    project.global_config["repository.pypi.password"] = "bar"
    repository = project.global_config.get_repository_config("pypi", "repository")
    assert repository.url == "https://upload.pypi.org/legacy/"
    assert repository.username == "foo"
    assert repository.password == "bar"

    project.global_config["repository.pypi.url"] = "https://example.pypi.org/legacy/"
    repository = project.global_config.get_repository_config("pypi", "repository")
    assert repository.url == "https://example.pypi.org/legacy/"


def test_hide_password_in_output_repository(project, pdm):
    assert project.global_config["repository.pypi.password"] is None
    project.global_config["repository.pypi.username"] = "testuser"
    project.global_config["repository.pypi.password"] = "secret"
    result = pdm(["config", "repository.pypi"], obj=project, strict=True)
    assert "password = <hidden>" in result.output
    result = pdm(["config", "repository.pypi.password"], obj=project, strict=True)
    assert "<hidden>" == result.output.strip()


def test_hide_password_in_output_pypi(project, pdm):
    with pytest.raises(KeyError):
        assert project.global_config["pypi.extra.password"] is None
    project.global_config["pypi.extra.username"] = "testuser"
    project.global_config["pypi.extra.password"] = "secret"
    project.global_config["pypi.extra.url"] = "https://test/simple"
    result = pdm(["config", "pypi.extra"], obj=project, strict=True)
    assert "password = <hidden>" in result.output
    result = pdm(["config", "pypi.extra.password"], obj=project, strict=True)
    assert "<hidden>" == result.output.strip()
    result = pdm(["config"], obj=project)
    assert "pypi.extra.password" in result.output
    assert "<hidden>" in result.output


def test_config_get_repository(project, pdm):
    config = project.global_config["repository.pypi"]
    assert config == project.global_config.get_repository_config("pypi", "repository")
    assert project.global_config["repository.pypi.url"] == "https://upload.pypi.org/legacy/"

    result = pdm(["config", "repository.pypi"], obj=project, strict=True)
    assert result.stdout.strip() == "repository.pypi.url = https://upload.pypi.org/legacy/"

    assert (
        project.global_config.get_repository_config("https://example.pypi.org/legacy/", "repository").url
        == "https://example.pypi.org/legacy/"
    )

    result = pdm(["config", "repository.pypi.url"], obj=project, strict=True)
    assert result.stdout.strip() == "https://upload.pypi.org/legacy/"


def test_config_set_repository(project):
    project.global_config["repository.pypi.url"] = "https://example.pypi.org/legacy/"
    project.global_config["repository.pypi.username"] = "foo"
    assert project.global_config["repository.pypi.url"] == "https://example.pypi.org/legacy/"
    assert project.global_config["repository.pypi.username"] == "foo"
    del project.global_config["repository.pypi.username"]
    assert project.global_config["repository.pypi.username"] is None


def test_config_del_repository(project):
    project.global_config["repository.test.url"] = "https://example.org/simple"
    assert project.global_config.get_repository_config("test", "repository") is not None

    del project.global_config["repository.test"]
    assert project.global_config.get_repository_config("test", "repository") is None


def test_config_password_save_into_keyring(project, keyring):
    project.global_config.update(
        {
            "pypi.extra.url": "https://extra.pypi.org/simple",
            "pypi.extra.username": "foo",
            "pypi.extra.password": "barbaz",
            "repository.pypi.username": "frost",
            "repository.pypi.password": "password",
        }
    )

    assert project.global_config["pypi.extra.password"] == "barbaz"
    assert project.global_config["repository.pypi.password"] == "password"
    for key in ("pypi.extra", "repository.pypi"):
        assert "password" not in project.global_config._file_data[key]

    assert keyring.enabled
    assert keyring.get_auth_info("pdm-pypi-extra", "foo") == ("foo", "barbaz")
    assert keyring.get_auth_info("pdm-repository-pypi", None) == ("frost", "password")

    del project.global_config["pypi.extra"]
    del project.global_config["repository.pypi.password"]
    keyring.get_auth_info.cache_clear()
    assert keyring.get_auth_info("pdm-pypi-extra", "foo") is None
    assert keyring.get_auth_info("pdm-repository-pypi", None) is None


def test_keyring_operation_error_disables_itself(project, keyring, mocker):
    saver = mocker.patch.object(keyring.provider, "save_auth_info", side_effect=RuntimeError())
    getter = mocker.patch.object(keyring.provider, "get_auth_info")
    project.global_config.update(
        {
            "pypi.extra.url": "https://extra.pypi.org/simple",
            "pypi.extra.username": "foo",
            "pypi.extra.password": "barbaz",
            "repository.pypi.username": "frost",
            "repository.pypi.password": "password",
        }
    )

    assert project.global_config["pypi.extra.password"] == "barbaz"
    assert project.global_config["repository.pypi.password"] == "password"

    saver.assert_called_once()
    getter.assert_not_called()

    assert not keyring.enabled
    assert keyring.get_auth_info("pdm-pypi-extra", "foo") is None
    assert keyring.get_auth_info("pdm-repository-pypi", None) is None


def _symlink_or_skip(link, target):
    try:
        link.symlink_to(target)
    except OSError as e:
        pytest.skip(f"symlink is not supported: {e}")


def test_config_keeps_symlink_path(tmp_path):
    target = tmp_path / "target.toml"
    target.write_text("python.use_pyenv = false\n")
    link = tmp_path / "config.toml"
    _symlink_or_skip(link, target)

    config = Config(link)

    # The symlink path is kept (only made absolute), not resolved to the target.
    assert config.config_file == link.absolute()
    assert config.config_file.is_symlink()
    assert config["python.use_pyenv"] is False


def test_config_refuses_to_write_symlinked_file(tmp_path):
    target = tmp_path / "target.toml"
    target.write_text("python.use_pyenv = true\n")
    link = tmp_path / "config.toml"
    _symlink_or_skip(link, target)

    config = Config(link)

    with pytest.raises(PdmUsageError, match="symlink"):
        config["python.use_pyenv"] = False

    # The symlink target outside the intended file is left untouched.
    assert target.read_text() == "python.use_pyenv = true\n"
    assert link.is_symlink()


def test_global_config_writes_through_symlink(tmp_path):
    target = tmp_path / "target.toml"
    target.write_text("python.use_pyenv = true\n")
    link = tmp_path / "config.toml"
    _symlink_or_skip(link, target)

    # The global config is the user's own file; a user-managed symlink (e.g. a
    # dotfiles-managed config) is honored rather than refused.
    config = Config(link, is_global=True)
    config["python.use_pyenv"] = False

    assert link.is_symlink()
    assert config["python.use_pyenv"] is False
    assert "use_pyenv = false" in target.read_text()



import itertools

from pdm.resolver.graph import OrderedSet


def test_ordered_set():
    elems = ["A", "bb", "c3"]
    all_sets = set()
    for case in itertools.permutations(elems):
        s = OrderedSet(case)
        all_sets.add(s)
        assert list(s) == list(case)
        assert len(s) == len(case)
        for e in elems:
            assert e in s
            assert e + "1" not in s
        assert str(s) == f"{{{', '.join(map(repr, case))}}}"
        assert repr(s) == f"OrderedSet({{{', '.join(map(repr, case))}}})"

    assert len(all_sets) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pdm.project.core import Project


def test_session_sources_all_proxy(project: Project, mocker, monkeypatch):
    for key in ("http_proxy", "https_proxy", "HTTP_PROXY", "HTTPS_PROXY", "ALL_PROXY", "no_proxy", "NO_PROXY"):
        monkeypatch.delenv(key, raising=False)
    monkeypatch.setenv("all_proxy", "http://localhost:8888")
    mock_get_transport = mocker.patch("pdm.models.session._get_transport")

    assert project.environment.session is not None
    transport_args = mock_get_transport.call_args
    assert transport_args is not None
    assert transport_args.kwargs["proxy"].url == "http://localhost:8888"

    monkeypatch.setenv("no_proxy", "pypi.org")
    mock_get_transport.reset_mock()
    del project.environment.session
    assert project.environment.session is not None
    transport_args = mock_get_transport.call_args
    assert transport_args is not None
    assert transport_args.kwargs["proxy"] is None
